import copy

FIELD_SEPARATOR = '.'


class DataNode:

    def __init__(self, defaults={}):
        self.node = copy.deepcopy(defaults)

    def __setitem__(self, key, value):
        field_chain = key.split(FIELD_SEPARATOR)
        target_node = self._determine_node(field_chain)
        target_node[field_chain[-1]] = value

    def _determine_node(self, field_chain):
        current_node = self.node
        for field in field_chain[:len(field_chain) - 1]:
            if field not in current_node:
                current_node[field] = {}
            current_node = current_node[field]
        return current_node

    def __getitem__(self, key):
        field_chain = key.split(FIELD_SEPARATOR)
        current_node = self.node.get(field_chain[0])
        for field in field_chain[1:]:
            if current_node is None:
                break
            else:
                current_node = current_node.get(field)
        return current_node

    def as_dict(self):
        return copy.deepcopy(self.node)



#!/usr/bin/env python
"""
Description goes here
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "08/05/2018"

from ingest.template.spreadsheet_builder import SpreadsheetBuilder


# build a generic spreadsheet from the latest schemas
spreadsheet_builder = SpreadsheetBuilder("generic.xlsx")
spreadsheet_builder.generate_workbook()
spreadsheet_builder.save_workbook()


# build a spreadsheet given new dev schemas

dev_schemas = [
    "http://schema.dev.data.humancellatlas.org/type/project/latest/project",
    "http://schema.dev.data.humancellatlas.org/type/biomaterial/latest/cell_suspension",
    "http://schema.dev.data.humancellatlas.org/type/biomaterial/latest/specimen_from_organism",
    "http://schema.dev.data.humancellatlas.org/type/biomaterial/latest/donor_organism",
    "http://schema.dev.data.humancellatlas.org/type/file/latest/sequence_file",
    "http://schema.dev.data.humancellatlas.org/type/protocol/biomaterial_collection/latest/dissociation_protocol",
    "http://schema.dev.data.humancellatlas.org/type/protocol/biomaterial_collection/latest/enrichment_protocol",
    "http://schema.dev.data.humancellatlas.org/type/protocol/sequencing/latest/library_preparation_protocol",
    "http://schema.dev.data.humancellatlas.org/type/protocol/sequencing/latest/sequencing_protocol",
]

spreadsheet_builder = SpreadsheetBuilder("latest_dev_template.xlsx")
spreadsheet_builder.generate_workbook(schema_urls=dev_schemas)
spreadsheet_builder.save_workbook()

# build a spreadsheet given a spreadsheet layout given a set of schemas


dev_schemas.append("http://schema.dev.data.humancellatlas.org/type/process/latest/process")

spreadsheet_builder = SpreadsheetBuilder("my_tabs_config.xlsx")
spreadsheet_builder.generate_workbook(tabs_template="my_tabs_config.yaml", schema_urls=dev_schemas)
spreadsheet_builder.save_workbook()



from openpyxl import Workbook

SCHEMAS_WORKSHEET = 'Schemas'
PROJECT_WORKSHEET = 'Project'
CONTACT_WORKSHEET = 'Contact'
FUNDER_WORKSHEET = 'Funder'
PUBLICATION_WORKSHEET = 'Publications'

# TODO think of a better name
SPECIAL_TABS = [SCHEMAS_WORKSHEET, PROJECT_WORKSHEET]

# TODO figure out how to identify that a tab is a module tab to generate this config automatically
MODULE_TABS = {
    'Contact': {
        'field': 'contributors',
        'parent_entity': 'project'
    },
    'Funder': {
        'field': 'funders',
        'parent_entity': 'project'
    },
    'Publications': {
        'field': 'publications',
        'parent_entity': 'project'
    },
}


class IngestWorkbook:

    def __init__(self, workbook: Workbook):
        self.workbook = workbook

    def get_project_worksheet(self):
        return self.get_worksheet(PROJECT_WORKSHEET)

    def get_worksheet(self, worksheet_title):
        if worksheet_title in self.workbook.get_sheet_names():
            return self.workbook[worksheet_title]

        if worksheet_title.lower() in self.workbook.get_sheet_names():
            return self.workbook[worksheet_title.lower()]

        return None

    def get_schemas(self):
        schemas = []

        worksheet = None

        if SCHEMAS_WORKSHEET in self.workbook.get_sheet_names():
            worksheet = self.workbook.get_sheet_by_name(SCHEMAS_WORKSHEET)

        if SCHEMAS_WORKSHEET.lower() in self.workbook.get_sheet_names():
            worksheet = self.workbook.get_sheet_by_name(SCHEMAS_WORKSHEET.lower())

        if not worksheet:
            return schemas

        for row in worksheet.iter_rows(row_offset=1, max_row=(worksheet.max_row - 1)):
            schema_cell = row[0]
            schemas.append(schema_cell.value)
        return schemas

    def importable_worksheets(self):
        importable_names = [name for name in self.workbook.get_sheet_names() if
                            (not name in SPECIAL_TABS and not name in list(MODULE_TABS.keys()))]
        return [self.workbook[name] for name in importable_names]

    def module_worksheets(self):
        return [self.get_worksheet(name) for name in list(MODULE_TABS.keys())]

    def get_module_field(self, module_tab_name):
        return MODULE_TABS[module_tab_name]['field'] if MODULE_TABS.get(module_tab_name) and \
                                                        MODULE_TABS[module_tab_name].get('field') else None




import json
from unittest import TestCase

import copy
from mock import MagicMock, patch, call

from ingest.api.ingestapi import IngestApi
from ingest.importer.data_node import DataNode
from ingest.importer.submission import Submission, Entity, IngestSubmitter, EntityLinker, LinkedEntityNotFound, \
    InvalidLinkInSpreadsheet, MultipleProcessesFound, EntityMap

import ingest.api.ingestapi


class SubmissionTest(TestCase):

    def test_new_submission(self):
        # given
        ingest.api.ingestapi.requests.get = MagicMock()
        mock_ingest_api = MagicMock(name='ingest_api')

        submission = Submission(mock_ingest_api, submission_url='submission_url')
        submission_url = submission.get_submission_url()

        self.assertEqual('submission_url', submission_url)

    def test_add_metadata(self):
        new_entity_mock_response = {
            'content': {},
            'submissionDate': '2018-05-08T10:17:49.476Z',
            'updateDate': '2018-05-08T10:17:57.254Z',
            'user': 'anonymousUser',
            'lastModifiedUser': 'anonymousUser',
            'uuid': {
                'uuid': '5a36689b-302b-40e4-bef1-837b47f0cb51'
            },
            'validationState': 'Draft',
            '_links': {
                'self': {
                    'href': 'http://api.ingest.dev.data.humancellatlas.org/protocols/5af1794d6a65a50007755b6a'
                },
                'protocol': {
                    'href': 'http://api.ingest.dev.data.humancellatlas.org/protocols/5af1794d6a65a50007755b6a',
                    'title': 'A single protocol'
                },
                'submissionEnvelopes': {
                    'href': 'http://api.ingest.dev.data.humancellatlas.org/protocols/5af1794d6a65a50007755b6a/submissionEnvelopes',
                    'title': 'Access or create new submission envelopes'
                }
            }
        }

        ingest.api.ingestapi.requests.get = MagicMock()
        mock_ingest_api = MagicMock(name='ingest_api')
        mock_ingest_api.load_root = MagicMock()
        mock_ingest_api.createEntity = MagicMock(return_value=new_entity_mock_response)

        submission = Submission(mock_ingest_api, submission_url='url')

        entity = Entity(entity_id='id', entity_type='biomaterial', content={})
        entity = submission.add_entity(entity)

        self.assertEqual(new_entity_mock_response, entity.ingest_json)

    def test_define_manifest(self):
        # expect:
        self._do_test_define_manifest(32)
        self._do_test_define_manifest(899)
        self._do_test_define_manifest(45219)

    def _do_test_define_manifest(self, total_count):
        # given:
        ingest_api = MagicMock('ingest_api')
        ingest_api.createSubmissionManifest = MagicMock()
        url = 'http://core.sample.com/submission/8fd733'
        submission = Submission(ingest_api, url)

        count_per_entity = {
            'biomaterial': 5,
            'project': 1,
            'protocol': 5,
            'file': 10,
            'process': 5
        }
        # and:
        entity_map = MagicMock('entity_map')
        entity_map.count_total = MagicMock(return_value=total_count)
        entity_map.count_entities_of_type = lambda entity: count_per_entity.get(entity)

        # when:
        submission.define_manifest(entity_map)

        # then:
        ingest_api.createSubmissionManifest.assert_called()
        ingest_api_args, __ = ingest_api.createSubmissionManifest.call_args
        self.assertEqual(2, len(ingest_api_args))
        self.assertEqual(url, ingest_api_args[0])

        # and:
        raw_json = ingest_api_args[1]
        submitted_json = json.loads(raw_json)
        self.assertEqual(total_count, submitted_json['totalCount'])
        self.assertEqual(count_per_entity['biomaterial'], submitted_json['expectedBiomaterials'])
        self.assertEqual(count_per_entity['process'], submitted_json['expectedProcesses'])
        self.assertEqual(count_per_entity['file'], submitted_json['expectedFiles'])
        self.assertEqual(count_per_entity['protocol'], submitted_json['expectedProtocols'])
        self.assertEqual(count_per_entity['project'], submitted_json['expectedProjects'])


def _create_spreadsheet_json():
    spreadsheet_json = {
        'project': {
            'dummy-project-id': {
                'content': {
                    'key': 'project_1'
                }
            }
        },
        'biomaterial': {
            'biomaterial_id_1': {
                'content': {
                    'key': 'biomaterial_1'
                }
            },
            'biomaterial_id_2': {
                'content': {
                    'key': 'biomaterial_2'
                },
                'links_by_entity': {
                    'biomaterial': ['biomaterial_id_1'],
                    'process': ['process_id_1']
                }
            },
            'biomaterial_id_3': {
                'content': {
                    'key': 'biomaterial_3'
                },
                'links_by_entity': {
                    'biomaterial': ['biomaterial_id_2'],
                    'process': ['process_id_2']
                }
            },
        },
        'file': {
            'file_id_1': {
                'content': {
                    'file_core': {
                        'file_name': 'file_name'
                    }
                },
                'links_by_entity': {
                    'biomaterial': ['biomaterial_id_3']
                }
            }
        },
        'protocol': {
            'protocol_id_1': {
                'content': {
                    'key': 'protocol_1'
                }
            }
        }
    }

    return spreadsheet_json


class IngestSubmitterTest(TestCase):

    @patch('ingest.importer.submission.Submission')
    def test_submit(self, submission_constructor):
        # given:
        ingest_api = MagicMock('ingest_api')
        ingest_api.getSubmissionEnvelope = MagicMock()
        submission = self._mock_submission(submission_constructor)

        # and:
        product = Entity('product', 'product_1', {})
        project = Entity('project', 'id', {})
        user = Entity('user', 'user_1', {})
        entity_map = EntityMap(product, user, project)


        # when:
        submitter = IngestSubmitter(ingest_api)
        submitter.submit(entity_map, submission_url='url')

        # then:
        submission_constructor.assert_called_with(ingest_api, 'url')
        submission.define_manifest.assert_called_with(entity_map)
        submission.add_entity.assert_has_calls([call(product), call(user)], any_order=True)

    @patch('ingest.importer.submission.Submission')
    def test_submit_linked_entity(self, submission_constructor):
        # given:
        ingest_api = MagicMock('ingest_api')
        ingest_api.getSubmissionEnvelope = MagicMock()
        submission = self._mock_submission(submission_constructor)

        # and:
        user = Entity('user', 'user_1', {})
        entity_map = EntityMap(user)

        # and:
        link_to_user = {
            'entity': 'user',
            'id': 'user_1',
            'relationship': 'wish_list'
        }
        linked_product = Entity('product', 'product_1', {}, direct_links=[link_to_user])
        project = Entity('project', 'id', {})
        entity_map.add_entity(linked_product)
        entity_map.add_entity(project)

        # when:
        submitter = IngestSubmitter(ingest_api)
        submitter.submit(entity_map, submission_url='url')

        # then:
        submission_constructor.assert_called_with(ingest_api, 'url')
        submission.define_manifest.assert_called_with(entity_map)
        submission.add_entity.assert_has_calls([call(user), call(linked_product)], any_order=True)
        submission.link_entity.assert_called_with(linked_product, user, relationship='wish_list')

    @staticmethod
    def _mock_submission(submission_constructor):
        submission = MagicMock('submission')
        submission.define_manifest = MagicMock()
        submission.add_entity = MagicMock()
        submission.link_entity = MagicMock()
        submission_constructor.return_value = submission
        return submission


class EntityMapTest(TestCase):

    def test_load(self):
        # given:
        spreadsheet_json = _create_spreadsheet_json()

        # when:
        entity_map = EntityMap.load(spreadsheet_json)

        # then:
        self.assertEqual(['project', 'biomaterial', 'file', 'protocol'],
                         list(entity_map.get_entity_types()))

        # and:
        # TODO shouldn't entity id's be unique and that there's no need to specify entity type?
        biomaterial1 = entity_map.get_entity('biomaterial', 'biomaterial_id_1')
        self._assert_correct_entity(biomaterial1, entity_id='biomaterial_id_1',
                                    entity_type='biomaterial', content={'key': 'biomaterial_1'})

        # and:
        biomaterial2 = entity_map.get_entity('biomaterial', 'biomaterial_id_2')
        links = {'biomaterial': ['biomaterial_id_1'], 'process': ['process_id_1']}
        self._assert_correct_entity(biomaterial2, entity_id='biomaterial_id_2',
                                    entity_type='biomaterial', content={'key': 'biomaterial_2'},
                                    links=links)

        # and:
        protocol1 = entity_map.get_entity('protocol', 'protocol_id_1')
        self.assertEqual({'key': 'protocol_1'}, protocol1.content)

    def _assert_correct_entity(self, entity, entity_id='', content={}, entity_type='', links={}):
        self.assertTrue(entity)
        self.assertEqual(entity_id, entity.id)
        self.assertEqual(content, entity.content)
        self.assertEqual(entity_type, entity.type)
        self.assertEqual(links, entity.links_by_entity)

    def test_count_total(self):
        # given:
        zero_map = EntityMap()

        # and:
        one_map = EntityMap()
        one_map.add_entity(Entity('product', 'product_1', {}))

        # and:
        three_map = EntityMap()
        three_map.add_entity(Entity('profile', 'profile_1', {}))
        for product_id in range(0, 2):
            three_map.add_entity(Entity('product', f'product_{product_id}', {}))

        # expect:
        self.assertEqual(0, zero_map.count_total())
        self.assertEqual(1, one_map.count_total())
        self.assertEqual(3, three_map.count_total())


class EntityLinkerTest(TestCase):

    def setUp(self):
        mocked_template_manager = MagicMock(name='template_manager')
        mocked_template_manager.get_schema_url = MagicMock(return_value='url')
        self.mocked_template_manager = mocked_template_manager

    def test_generate_direct_links_biomaterial_to_biomaterial_has_process(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    }
                },
                'biomaterial_id_2': {
                    'content': {
                        'key': 'biomaterial_2'
                    },
                    'links_by_entity': {
                        'biomaterial': ['biomaterial_id_1'],
                        'process': ['process_id_1'],
                        'protocol': ['protocol_id_1', 'protocol_id_2']
                    }

                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    }
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    }
                }
            }
        }

        expected_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'inputToProcesses'
                        }
                    ]
                },
                'biomaterial_id_2': {
                    'content': {
                        'key': 'biomaterial_2'
                    },
                    'links_by_entity': {
                        'biomaterial': ['biomaterial_id_1'],
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'derivedByProcesses'
                        }
                    ]
                }
            },
            'process': {
                'process_id_1': {
                    'content': {
                        'key': 'process_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_1',
                            'relationship': 'protocols'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_2',
                            'relationship': 'protocols'
                        }
                    ]
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    },
                    'direct_links': [
                    ]
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    },
                    'direct_links': [
                    ]
                }
            }
        }

        entity_linker = EntityLinker(self.mocked_template_manager)
        entities_dictionaries = EntityMap.load(spreadsheet_json)
        output = entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self._assert_equal_direct_links(expected_json, output)

    def _assert_equal_direct_links(self, expected_json, output):
        for entity_type, entities_dict in expected_json.items():
            for entity_id, entity_dict in entities_dict.items():
                expected_links = entities_dict[entity_id].get('direct_links')
                expected_links = expected_links if expected_links else []
                entity = output.get_entity(entity_type, entity_id)
                self.assertTrue(entity)

                for link in expected_links:
                    self.assertTrue(link in entity.direct_links, f'{json.dumps(link)} is not in direct links')

    def test_generate_direct_links_biomaterial_to_biomaterial_no_process(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    }
                },
                'biomaterial_id_2': {
                    'content': {
                        'key': 'biomaterial_2'
                    },
                    'links_by_entity': {
                        'biomaterial': ['biomaterial_id_1'],
                        'protocol': ['protocol_id_1', 'protocol_id_2']
                    }

                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    }
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    }
                }
            }
        }

        expected_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'inputToProcesses'
                        }
                    ]
                },
                'biomaterial_id_2': {
                    'content': {
                        'key': 'biomaterial_2'
                    },
                    'links_by_entity': {
                        'biomaterial': ['biomaterial_id_1'],
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'derivedByProcesses'
                        }
                    ]
                }
            },
            'process': {
                'process_id_1': {
                    'content': {
                        'key': 'process_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_1',
                            'relationship': 'protocols'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_2',
                            'relationship': 'protocols'
                        }
                    ]
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    },
                    'direct_links': [
                    ]
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    },
                    'direct_links': [
                    ]
                }
            }
        }

        entity_linker = EntityLinker(self.mocked_template_manager)
        entities_dictionaries = EntityMap.load(spreadsheet_json)
        output = entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self._assert_equal_direct_links(expected_json, output)

    def test_generate_direct_links_file_to_file_no_process(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    }
                },
                'file_id_2': {
                    'content': {
                        'key': 'file_2'
                    },
                    'links_by_entity': {
                        'file': ['file_id_1'],
                        'protocol': ['protocol_id_1', 'protocol_id_2']
                    }
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    }
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    }
                }
            }
        }

        expected_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'inputToProcesses'
                        }
                    ]
                },
                'file_id_2': {
                    'content': {
                        'key': 'file_2'
                    },
                    'links_by_entity': {
                        'file': ['file_id_1'],
                    },
                    'direct_links': [
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'derivedByProcesses'
                        }
                    ]
                }
            },
            'process': {
                'process_id_1': {
                    'content': {
                        'key': 'process_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_1',
                            'relationship': 'protocols'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_2',
                            'relationship': 'protocols'
                        }
                    ]
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    },
                    'direct_links': [
                    ]
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    },
                    'direct_links': [
                    ]
                }
            }
        }

        entity_linker = EntityLinker(self.mocked_template_manager)
        entities_dictionaries = EntityMap.load(spreadsheet_json)
        output = entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self._assert_equal_direct_links(expected_json, output)

    def test_generate_direct_links_file_to_file_has_process(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    }
                },
                'file_id_2': {
                    'content': {
                        'key': 'file_2'
                    },
                    'links_by_entity': {
                        'file': ['file_id_1'],
                        'process': ['process_id_1'],
                        'protocol': ['protocol_id_1', 'protocol_id_2']
                    }

                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    }
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    }
                }
            }
        }

        expected_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'inputToProcesses'
                        }
                    ]
                },
                'file_id_2': {
                    'content': {
                        'key': 'file_2'
                    },
                    'links_by_entity': {
                        'file': ['file_id_1'],
                    },
                    'direct_links': [
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'derivedByProcesses'
                        },
                    ]
                }
            },
            'process': {
                'process_id_1': {
                    'content': {
                        'key': 'process_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_1',
                            'relationship': 'protocols'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_2',
                            'relationship': 'protocols'
                        }
                    ]
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    },
                    'direct_links': [
                    ]
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    },
                    'direct_links': [
                    ]
                }
            }
        }

        entity_linker = EntityLinker(self.mocked_template_manager)
        entities_dictionaries = EntityMap.load(spreadsheet_json)
        output = entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self._assert_equal_direct_links(expected_json, output)

    def test_generate_direct_links_file_to_biomaterial_has_process(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    },
                    'links_by_entity': {
                        'biomaterial': ['biomaterial_id_1'],
                        'process': ['process_id_1'],
                        'protocol': ['protocol_id_1', 'protocol_id_2']
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    }
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    }
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    }
                }
            }
        }

        expected_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'inputToProcesses'
                        }
                    ]
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'process',
                            'id': 'process_id_1',
                            'relationship': 'derivedByProcesses'
                        }
                    ]
                },
            },
            'process': {
                'process_id_1': {
                    'content': {
                        'key': 'process_1'
                    },
                    'direct_links': [
                        {
                            'entity': 'project',
                            'id': 'dummy-project-id',
                            'relationship': 'projects'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_1',
                            'relationship': 'protocols'
                        },
                        {
                            'entity': 'protocol',
                            'id': 'protocol_id_2',
                            'relationship': 'protocols'
                        }
                    ]
                }
            },
            'protocol': {
                'protocol_id_1': {
                    'content': {
                        'key': 'protocol_1'
                    },
                    'direct_links': [
                    ]
                },
                'protocol_id_2': {
                    'content': {
                        'key': 'protocol_2'
                    },
                    'direct_links': [
                    ]
                }
            }
        }

        entity_linker = EntityLinker(self.mocked_template_manager)
        entities_dictionaries = EntityMap.load(spreadsheet_json)
        output = entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self._assert_equal_direct_links(expected_json, output)

    def test_generate_direct_links_link_not_found_error(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'file': {
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    },
                    'links_by_entity': {
                        'biomaterial': ['biomaterial_id_1'],
                        'protocol': ['protocol_id_1', 'protocol_id_2']
                    }
                }
            }
        }

        entities_dictionaries = EntityMap.load(spreadsheet_json)
        entity_linker = EntityLinker(self.mocked_template_manager)

        with self.assertRaises(LinkedEntityNotFound) as context:
            entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self.assertEqual('biomaterial', context.exception.entity)
        self.assertEqual('biomaterial_id_1', context.exception.id)

    def test_generate_direct_links_invalid_spreadsheet_link(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    },
                    'links_by_entity': {
                        'file': ['file_id_1']
                    }
                }
            },
            'file':{
                'file_id_1': {
                    'content': {
                        'key': 'file_1'
                    }
                }
            }
        }

        entities_dictionaries = EntityMap.load(spreadsheet_json)
        entity_linker = EntityLinker(self.mocked_template_manager)

        with self.assertRaises(InvalidLinkInSpreadsheet) as context:
            entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self.assertEqual('biomaterial', context.exception.from_entity.type)
        self.assertEqual('file', context.exception.link_entity_type)

        self.assertEqual('biomaterial_id_1', context.exception.from_entity.id)
        self.assertEqual('file_id_1', context.exception.link_entity_id)

    def test_generate_direct_links_multiple_process_links(self):
        # given
        spreadsheet_json = {
            'project': {
                'dummy-project-id': {
                    'content': {
                        'key': 'project_1'
                    }
                }
            },
            'biomaterial': {
                'biomaterial_id_1': {
                    'content': {
                        'key': 'biomaterial_1'
                    },
                    'links_by_entity': {
                        'process': ['process_id_1', 'process_id_2']
                    }
                }
            },
            'process': {
                'process_id_1': {
                    'content': {
                        'key': 'process_1'
                    }
                },
                'process_id_2': {
                    'content': {
                        'key': 'process_2'
                    }
                }
            }
        }

        entities_dictionaries = EntityMap.load(spreadsheet_json)
        entity_linker = EntityLinker(self.mocked_template_manager)

        with self.assertRaises(MultipleProcessesFound) as context:
            entity_linker.process_links_from_spreadsheet(entities_dictionaries)

        self.assertEqual('biomaterial', context.exception.from_entity.type)
        self.assertEqual(['process_id_1', 'process_id_2'], context.exception.process_ids)



#!/usr/bin/env python
"""
Description goes here
"""

__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "01/05/2018"

import unittest
from unittest import TestCase
from unittest.mock import patch, MagicMock

import tests.template.schema_mock_utils as schema_mock

from ingest.template.schema_template import SchemaParser
from ingest.template.schema_template import SchemaTemplate
from ingest.template.schema_template import UnknownKeyException
from ingest.template.schema_template import RootSchemaException

class TestSchemaTemplate(TestCase):
    def setUp(self):
        self.longMessage = True
        self.dummyProjectUri = "https://schema.humancellatlas.org/type/project/5.1.0/project"
        self.dummyDonorUri = "https://schema.humancellatlas.org/type/biomaterial/5.1.0/donor_organism"
        pass

    def test_schema_lookup(self):

        data='{"id" : "'+self.dummyProjectUri+'", "properties": {"foo": "bar"} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertEqual("project", template.lookup('project.schema.domain_entity'))
        self.assertEqual("project", template.lookup('project.schema.module'))
        self.assertEqual("type", template.lookup('project.schema.high_level_entity'))

    def test_no_root_schema(self):
        data = '{"properties": {"foo": "bar"} }'
        with self.assertRaises(RootSchemaException):
            schema_mock.get_template_for_json(data=data)


    def test_unknown_key_exception(self):
        data = '{"id" : "'+self.dummyProjectUri+'", "properties": {"foo": "bar"} }'
        template = schema_mock.get_template_for_json(data=data)
        with self.assertRaises(UnknownKeyException):
                template.lookup('foo')

    def test_get_tab_name(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)

        tabs = template.get_tabs_config()
        self.assertEqual("donor_organism", tabs.get_key_for_label("donor_organism"))
        self.assertEqual("donor_organism", tabs.get_key_for_label("Donor organism"))

    def test_lookup_key_in_tab(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)

        self.assertEqual("donor_organism.foo_bar", template.get_key_for_label("Foo bar", "Donor organism"))

        with self.assertRaises(UnknownKeyException):
            template.get_key_for_label("Bar foo", "Donor organism")

    def test_required_fields(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "required": ["foo_bar"], "properties": { "foo_bar": {"user_friendly" : "Foo bar"}, "bar_foo" : {}} }'
        template = schema_mock.get_template_for_json(data=data)

        self.assertTrue(template.lookup("donor_organism.foo_bar.required"))
        self.assertFalse(template.lookup("donor_organism.bar_foo.required"))

    def test_has_type(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": { "foo_bar": {"type" : "string"} } }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertEqual("string", template.lookup("donor_organism.foo_bar.value_type"))
        self.assertFalse(template.lookup("donor_organism.foo_bar.multivalue"))


    def test_has_type_list(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": { "foo_bar": {"type" : "array" , "items" : {"type": "string"}} } }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertEqual("string", template.lookup("donor_organism.foo_bar.value_type"))
        self.assertTrue(template.lookup("donor_organism.foo_bar.multivalue"))

    def test_identifiable(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"biomaterial_id": {"identifiable" : "true"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertTrue(template.lookup("donor_organism.biomaterial_id.identifiable"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertFalse(template.lookup("donor_organism.foo_bar.identifiable"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"process_id": {"identifiable" : "true"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertTrue(template.lookup("donor_organism.process_id.identifiable"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"protocol_id": {"identifiable" : "true"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertTrue(template.lookup("donor_organism.protocol_id.identifiable"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"file_name": {"identifiable" : "true"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertTrue(template.lookup("donor_organism.file_name.identifiable"))

    def test_get_key_for_label(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)

        tabs = template.get_tabs_config()
        self.assertEqual("donor_organism", tabs.get_key_for_label("donor_organism"))

    def test_description(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"description" : "Foo is a bar"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertEqual("Foo is a bar", template.lookup("donor_organism.foo_bar.description"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertIsNone(template.lookup("donor_organism.foo_bar.description"))

    def test_format(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"format" : "date-time"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertEqual("date-time", template.lookup("donor_organism.foo_bar.format"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertIsNone(template.lookup("donor_organism.foo_bar.format"))

    def test_retrieveable(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"format" : "date-time"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertTrue(template.lookup("donor_organism.uuid.external_reference"))
        self.assertTrue(template.lookup("donor_organism.uuid.identifiable"))
        self.assertFalse(template.lookup("donor_organism.foo_bar.external_reference"))
        self.assertFalse(template.lookup("donor_organism.foo_bar.identifiable"))

        with self.assertRaises(UnknownKeyException):
            self.assertTrue(template.lookup("donor_organism.format.uuid.retrievable"))



    def test_example(self):
        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"example" : "Foo is a bar"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertEqual("Foo is a bar", template.lookup("donor_organism.foo_bar.example"))

        data = '{"id" : "' + self.dummyDonorUri + '", "properties": {"foo_bar": {"user_friendly" : "Foo bar"}} }'
        template = schema_mock.get_template_for_json(data=data)
        self.assertIsNone(template.lookup("donor_organism.foo_bar.example"))

    def test_follows_item_refs(self):
        data = '{"id" : "' + self.dummyProjectUri + '", "properties": { "foo_bar": {"type" : "array" , "items" : {"type" : "object", "id": "'+self.dummyDonorUri+'"}} } }'
        template = schema_mock.get_template_for_json(data=data)

        self.assertEqual("biomaterial", template.lookup("project.foo_bar.schema.domain_entity"))

    def test_get_domain_entity_from_url(self):
        schema_parser = SchemaParser(None)
        url = "https://schema.humancellatlas.org/type/project/5.1.0/project"
        self.assertEqual("project", schema_parser.get_domain_entity_from_url(url))
        url = "https://schema.humancellatlas.org/type/foo/bar/5.1.0/project"
        self.assertEqual("foo/bar", schema_parser.get_domain_entity_from_url(url))
        url = "https://schema.humancellatlas.org/type/foo/bar/latest/project"
        self.assertEqual("foo/bar", schema_parser.get_domain_entity_from_url(url))
        url = "https://schema.humancellatlas.org/type/foo/latest/project"
        self.assertEqual("foo", schema_parser.get_domain_entity_from_url(url))
        url = 'https://schema.humancellatlas.org/bundle/1.0.0/links'
        self.assertEqual(None, schema_parser.get_domain_entity_from_url(url))
        url= 'http://schema.dev.data.humancellatlas.org/system/1.0.0/provenance'
        self.assertEqual(None, schema_parser.get_domain_entity_from_url(url))

    def test_get_high_level_entity_from_url(self):
        schema_parser = SchemaParser(None)
        url = "https://schema.humancellatlas.org/type/project/5.1.0/project"
        self.assertEqual("type", schema_parser.get_high_level_entity_from_url(url))
        url = 'http://schema.dev.data.humancellatlas.org/system/1.0.0/provenance'
        self.assertEqual('system', schema_parser.get_high_level_entity_from_url(url))

    def test_get_module_from_url(self):
        schema_parser = SchemaParser(None)
        url = "https://schema.humancellatlas.org/type/project/5.1.0/project"
        self.assertEqual("project", schema_parser.get_module_from_url(url))

if __name__ == '__main__':
    unittest.main()


from unittest import TestCase

from openpyxl import Workbook

from ingest.importer.spreadsheet.ingest_workbook import IngestWorkbook


class IngestWorkbookTest(TestCase):

    def test_get_schemas(self):
        # given:
        workbook = Workbook()

        # and:
        schemas_sheet = workbook.create_sheet('Schemas')
        schemas_sheet['A1'] = 'schema'

        # and:
        base_url = 'https://schema.humancellatlas.org'
        expected_schemas = [
            f'{base_url}/type/biomaterial/cell_suspension',
            f'{base_url}type/biomaterial/organ_from_donor',
            f'{base_url}/type/process/library_preparation'
        ]
        header_offset = 2
        for index, schema in enumerate(expected_schemas):
            schemas_sheet[f'A{index + header_offset}'] = schema

        # and:
        ingest_workbook = IngestWorkbook(workbook)

        # when:
        actual_schemas = ingest_workbook.get_schemas()

        # then:
        self.assertEqual(expected_schemas, actual_schemas)

    def test_importable_worksheets(self):
        # given:
        workbook = Workbook()

        # and:
        importable_names = ['Organ From Donor', 'Cell Suspension']
        expected_worksheets = [workbook.create_sheet(name) for name in importable_names]
        workbook.create_sheet('Schemas')

        # and:
        default_worksheet = workbook.get_sheet_by_name('Sheet')
        workbook.remove_sheet(default_worksheet)

        # and:
        ingest_workbook = IngestWorkbook(workbook)

        # when:
        actual_worksheets = ingest_workbook.importable_worksheets()

        # then:
        self.assertEqual(expected_worksheets, actual_worksheets)



import copy
import logging

from openpyxl.worksheet import Worksheet

import ingest.template.schema_template as schema_template
from ingest.api.ingestapi import IngestApi
from ingest.importer.conversion import utils, conversion_strategy
from ingest.importer.conversion.column_specification import ColumnSpecification
from ingest.importer.conversion.conversion_strategy import CellConversion, \
    ListElementCellConversion, FieldOfSingleElementListCellConversion
from ingest.importer.conversion.metadata_entity import MetadataEntity
from ingest.importer.data_node import DataNode
from ingest.template.schema_template import SchemaTemplate


class TemplateManager:

    def __init__(self, template:SchemaTemplate, ingest_api:IngestApi):
        self.template = template
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)

    def create_template_node(self, worksheet: Worksheet):
        concrete_entity = self.get_concrete_entity_of_tab(worksheet.title)
        schema = self._get_schema(concrete_entity)
        data_node = DataNode()
        data_node['describedBy'] = schema['url']
        data_node['schema_type'] = schema['domain_entity']
        return data_node

    def create_row_template(self, worksheet: Worksheet):
        tab_name = worksheet.title
        object_type = self.get_concrete_entity_of_tab(tab_name)
        header_row = self.get_header_row(worksheet)
        cell_conversions = []

        header_counter = {}
        for cell in header_row:
            header = cell.value

            if not header_counter.get(header):
                header_counter[header] = 0
            header_counter[header] = header_counter[header] + 1

            column_spec = self._define_column_spec(header, object_type,
                                                   order_of_occurence=header_counter[header])
            strategy = conversion_strategy.determine_strategy(column_spec)
            cell_conversions.append(strategy)

        default_values = self._define_default_values(object_type)
        return RowTemplate(cell_conversions, default_values=default_values)

    def create_simple_row_template(self, worksheet: Worksheet):
        tab_name = worksheet.title
        object_type = self.get_concrete_entity_of_tab(tab_name)
        header_row = self.get_header_row(worksheet)

        cell_conversions = []
        for cell in header_row:
            header = cell.value
            column_spec = self._define_column_spec(header, object_type)
            strategy = FieldOfSingleElementListCellConversion(column_spec.field_name,
                                                 column_spec.determine_converter())
            cell_conversions.append(strategy)

        default_values = self._define_default_values(object_type)
        return RowTemplate(cell_conversions, default_values=default_values)

    # TODO move this outside template manager
    @staticmethod
    def get_header_row(worksheet):
        for row in worksheet.iter_rows(row_offset=3, max_row=1):
            header_row = row

        clean_header_row = []

        for cell in header_row:
            if cell.value is None:
                continue
            clean_header_row.append(cell)

        return clean_header_row

    def _define_column_spec(self, header, object_type, order_of_occurence=1):
        if header is not None:
            parent_path, __ = utils.split_field_chain(header)
            raw_spec = self.lookup(header)
            raw_parent_spec = self.lookup(parent_path)
            concrete_type = utils.extract_root_field(header)
            main_category = self.get_domain_entity(concrete_type)
            column_spec = ColumnSpecification.build_raw(header, object_type, main_category, raw_spec,
                                                        parent=raw_parent_spec,
                                                        order_of_occurence=order_of_occurence)
        else:
            column_spec = None
        return column_spec

    def _define_default_values(self, object_type):
        default_values = {
            'describedBy': self.get_schema_url(object_type),
            'schema_type': self.get_domain_entity(object_type)
        }
        return default_values

    def get_latest_schema_url(self, high_level_entity, domain_entity, concrete_entity):
        latest_schema = self.ingest_api.get_schemas(
            latest_only=True,
            high_level_entity=high_level_entity,
            domain_entity=domain_entity.split('/')[0],
            concrete_entity=concrete_entity
        )

        return latest_schema[0]['_links']['json-schema']['href'] if latest_schema else None

    def get_schema_url(self, concrete_entity):
        schema = self._get_schema(concrete_entity)
        # TODO use schema version that is specified in spreadsheet for now
        return schema.get('url') if schema else None

    def get_domain_entity(self, concrete_entity):
        domain_entity = None
        schema = self._get_schema(concrete_entity)
        if schema:
            domain_entity = schema.get('domain_entity', '')
            subdomain = domain_entity.split('/')
            if subdomain:
                domain_entity = subdomain[0]
        return domain_entity

    def _get_schema(self, concrete_entity):
        spec = self.lookup(concrete_entity)
        return spec.get('schema') if spec else None

    def get_concrete_entity_of_tab(self, tab_name):
        concrete_entity = None
        try:
            tabs_config = self.template.get_tabs_config()
            concrete_entity = tabs_config.get_key_for_label(tab_name)
        except schema_template.UnknownKeyException as e:
            pass
        except KeyError as e:
            pass

        return concrete_entity

    def get_key_for_label(self, header_name, tab_name):
        try:
            key = self.template.get_key_for_label(header_name, tab_name)
        except:
            self.logger.warning(f'{header_name} in "{tab_name}" tab is not found in schema template')
        return key

    def lookup(self, header_name):
        try:
            spec = self.template.lookup(header_name)
        except schema_template.UnknownKeyException:
            self.logger.warning(f'schema_template.UnknownKeyException: Could not lookup {header_name} in template.')
            return {}

        return spec


def build(schemas, ingest_api) -> TemplateManager:
    template = None

    if not schemas:
        template = SchemaTemplate(ingest_api_url=ingest_api.url)
    else:
        template = SchemaTemplate(ingest_api_url=ingest_api.url, list_of_schema_urls=schemas)

    template_mgr = TemplateManager(template, ingest_api)
    return template_mgr


class RowTemplate:

    def __init__(self, cell_conversions, default_values={}):
        self.cell_conversions = cell_conversions
        self.default_values = copy.deepcopy(default_values)

    def do_import(self, row):
        metadata = MetadataEntity(content=self.default_values)
        for index, cell in enumerate(row):
            if cell.value is None:
                continue
            conversion: CellConversion = self.cell_conversions[index]
            conversion.apply(metadata, cell.value)
        return metadata


class ParentFieldNotFound(Exception):
    def __init__(self, header_name):
        message = f'{header_name} has no parent field'
        super(ParentFieldNotFound, self).__init__(message)

        self.header_name = header_name




class InvalidBooleanValue(Exception):

    def __init__(self, value):
        self.value = value

    def get_value(self):
        return self.value


class UnknownMainCategory(Exception):

    DEFAULT_MESSAGE = 'Main category was not specified.'

    def __init__(self, message=DEFAULT_MESSAGE):
        self.message = message

    def __str__(self):
        return repr(self.message)



import os
import unittest
from unittest import TestCase

from mock import MagicMock, patch
from openpyxl import Workbook

from ingest.importer.conversion import conversion_strategy
from ingest.importer.conversion.metadata_entity import MetadataEntity
from ingest.importer.data_node import DataNode
from ingest.importer.importer import WorksheetImporter, WorkbookImporter, XlsImporter, IdentifiableWorksheetImporter
from ingest.importer.spreadsheet.ingest_workbook import IngestWorkbook

BASE_PATH = os.path.dirname(__file__)

HEADER_ROW = 4


def _create_single_row_worksheet(worksheet_data: dict):
    workbook = Workbook()
    worksheet = workbook.create_sheet()

    for column, data in worksheet_data.items():
        key, value = data
        worksheet[f'{column}{HEADER_ROW}'] = key
        worksheet[f'{column}6'] = value

    return worksheet


class WorkbookImporterTest(TestCase):

    @patch('ingest.importer.importer.IdentifiableWorksheetImporter')
    def test_do_import(self, worksheet_importer_constructor):
        # given: set up template manager
        key_label_map = {
            'Project': 'project',
            'Cell Suspension': 'cell_suspension'
        }

        domain_entity_map = {
            'project': 'project',
            'cell_suspension': 'biomaterial'
        }

        mock_template_manager = MagicMock()
        mock_template_manager.get_concrete_entity_of_tab = lambda key: key_label_map.get(key)
        mock_template_manager.get_domain_entity = lambda key: domain_entity_map.get(key)

        # and: set up worksheet importer
        worksheet_importer = IdentifiableWorksheetImporter()
        expected_json = self._fake_worksheet_import(worksheet_importer, mock_template_manager)

        # and: set up workbook
        workbook = Workbook()
        ingest_workbook = IngestWorkbook(workbook)
        schema_list = self._mock_get_schemas(ingest_workbook)
        self._mock_importable_worksheets(ingest_workbook, workbook)

        # and: mock WorksheetImporter constructor
        worksheet_importer_constructor.return_value = worksheet_importer
        workbook_importer = WorkbookImporter(mock_template_manager)
        workbook_importer.import_or_reference_project = MagicMock()

        # when:
        workbook_output = workbook_importer.do_import(ingest_workbook)

        # then:
        self.assertEqual(2, len(list(workbook_output.keys())))
        self.assertEqual(['project', 'biomaterial'], list(workbook_output.keys()))
        self.assertEqual(2, len(list(workbook_output['project'].keys())))
        self.assertEqual(expected_json['project'], workbook_output['project'])
        self.assertEqual(expected_json['biomaterial'], workbook_output['biomaterial'])

    def _mock_get_schemas(self, ingest_workbook):
        schema_base_url = 'https://schema.humancellatlas.org'
        schema_list = [
            f'{schema_base_url}/type/project',
            f'{schema_base_url}/type/biomaterial'
        ]
        ingest_workbook.get_schemas = MagicMock(return_value=schema_list)
        return schema_list

    def _mock_importable_worksheets(self, ingest_workbook, workbook):
        project_worksheet = workbook.create_sheet('Project')
        cell_suspension_worksheet = workbook.create_sheet('Cell Suspension')
        ingest_workbook.importable_worksheets = MagicMock(return_value=[
            project_worksheet, cell_suspension_worksheet
        ])

    def _fake_worksheet_import(self, worksheet_importer: WorksheetImporter, mock_template_manager):
        projects = {
            'project 1': {'short_name': 'project 1', 'description': 'first project'},
            'project 2': {'short_name': 'project 2', 'description': 'second project'}
        }

        cell_suspensions = {
            'cell_suspension_101': {'biomaterial_id': 'cell_suspension_101', 'biomaterial_name': 'cell suspension'}
        }

        worksheet_iterator = iter([projects, cell_suspensions])
        worksheet_importer.do_import = (
            lambda __, tm: worksheet_iterator.__next__() if tm is mock_template_manager else []
        )

        return {
            'project': projects,
            'biomaterial': cell_suspensions
        }


class WorksheetImporterTest(TestCase):

    def test_do_import(self):
        # given:
        row_template = MagicMock('row_template')

        # and:
        john_doe_content = {'name': 'John Doe'}
        john_doe_links = {}
        john_doe_external_links = {'organisations': ['org_88', 'org_110', 'org_452']}
        john_doe_linking_details = {'position': 'manager'}
        john_doe = MetadataEntity(object_id='profile_1', content=john_doe_content,
                                  links=john_doe_links, external_links=john_doe_external_links,
                                  linking_details=john_doe_linking_details)

        # and:
        emma_jackson_content = {'name': 'Emma Jackson'}
        emma_jackson_links = {'friends': ['profile_19', 'profile_8']}
        emma_jackson_external_links = {}
        emma_jackson_linking_details = {}
        emma_jackson = MetadataEntity(object_id='profile_2', content=emma_jackson_content,
                                      links=emma_jackson_links,
                                      external_links=emma_jackson_external_links,
                                      linking_details=emma_jackson_linking_details)

        # and:
        row_template.do_import = MagicMock('import_row', side_effect=[john_doe, emma_jackson])

        # and:
        mock_template_manager = MagicMock('template_manager')
        mock_template_manager.create_row_template = MagicMock(return_value=row_template)
        mock_template_manager.get_header_row = MagicMock(return_value=['header1', 'header2'])
        mock_template_manager.get_concrete_entity_of_tab = MagicMock(return_value='concrete_entity')

        # and:
        workbook = Workbook()
        worksheet = workbook.create_sheet('user_profile')
        worksheet['A6'] = 'john'
        worksheet['A7'] = 'emma'

        # when:
        worksheet_importer = WorksheetImporter()
        profile = worksheet_importer.do_import(worksheet, mock_template_manager)

        # then:
        self.assertEqual(2, len(profile.keys()))
        self._assert_correct_profile(profile, 'profile_1', john_doe_content, john_doe_links,
                                     john_doe_external_links, john_doe_linking_details)
        self._assert_correct_profile(profile, 'profile_2', emma_jackson_content,
                                     emma_jackson_links, emma_jackson_external_links,
                                     emma_jackson_linking_details)

    def test_do_import_no_id_metadata(self):
        # given:
        row_template = MagicMock('row_template')

        # and:
        paper_metadata = MetadataEntity(content={'product_name': 'paper'},
                                        links={'delivery': ['123', '456']})
        pen_metadata = MetadataEntity(content={'product_name': 'pen'},
                                      links={'delivery': ['789']})
        row_template.do_import = MagicMock(side_effect=[paper_metadata, pen_metadata])

        # and:
        mock_template_manager = MagicMock('template_manager')
        mock_template_manager.create_row_template = MagicMock(return_value=row_template)
        mock_template_manager.get_header_row = MagicMock(return_value=['header1', 'header2'])
        mock_template_manager.get_concrete_entity_of_tab = MagicMock(return_value='concrete_entity')

        # and:
        workbook = Workbook()
        worksheet = workbook.create_sheet('product')
        worksheet['A6'] = 'paper'
        worksheet['A7'] = 'pen'

        # when:
        worksheet_importer = WorksheetImporter()
        result = worksheet_importer.do_import(worksheet, mock_template_manager)

        # then:
        self.assertEqual(2, len(result.keys()))

    def _assert_correct_profile(self, profile, profile_id, expected_content, expected_links,
                                expected_external_links, expected_linking_details):
        actual_profile = profile.get(profile_id)
        self.assertIsNotNone(actual_profile)
        self.assertEqual(expected_content, actual_profile.get('content'))
        self.assertEqual(expected_links, actual_profile.get('links_by_entity'))
        self.assertEqual(expected_external_links, actual_profile.get('external_links_by_entity'))
        self.assertEqual(expected_linking_details, actual_profile.get('linking_details'))


class IngestImporterTest(TestCase):

    # TODO why is this hitting servers?
    @unittest.skip
    def test_import_spreadsheet(self):

        spreadsheet_file = BASE_PATH + '/metadata_spleen_new_protocols.xlsx'

        submission = XlsImporter(MagicMock()).import_file(file_path=spreadsheet_file, submission_url=None, dry_run=True)

        self.assertTrue(submission)



from unittest import TestCase

import copy

from ingest.importer.conversion.column_specification import ColumnSpecification, ConversionType
from ingest.importer.conversion.data_converter import DataType, IntegerConverter, \
    BooleanConverter, ListConverter, StringConverter, DefaultConverter


class ColumnSpecificationTest(TestCase):

    def test_build_raw_single_type(self):
        # given:
        raw_string_spec = {
            'value_type': 'string',
            'multivalue': False,
            'identifiable': True
        }

        # when:
        string_column_spec = ColumnSpecification.build_raw('user.name', 'user', 'profile_entry',
                                                           raw_string_spec)

        # then:
        self.assertEqual('user.name', string_column_spec.field_name)
        self.assertEqual('profile_entry', string_column_spec.main_category)
        self.assertEqual(DataType.STRING, string_column_spec.data_type)
        self.assertFalse(string_column_spec.is_multivalue())
        self.assertTrue(string_column_spec.is_identity())

    def test_build_raw_multivalue(self):
        # given:
        raw_int_array_spec = {
            'value_type': 'integer',
            'multivalue': True
        }

        # when:
        int_array_column_spec = ColumnSpecification.build_raw('sample.numbers', 'user',
                                                              'profile_entry', raw_int_array_spec)

        # then:
        self.assertEqual('sample.numbers', int_array_column_spec.field_name)
        self.assertEqual('profile_entry', int_array_column_spec.main_category)
        self.assertEqual(DataType.INTEGER, int_array_column_spec.data_type)
        self.assertTrue(int_array_column_spec.is_multivalue())
        self.assertFalse(int_array_column_spec.is_identity())

    def test_build_raw_external_reference(self):
        # given:
        external_raw_spec = {
            # 'value_type': 'string',
            'identifiable': True,
            'external_reference': True
        }

        # and:
        non_external_raw_spec = copy.deepcopy(external_raw_spec)
        non_external_raw_spec['external_reference'] = False

        # when:
        external_spec = ColumnSpecification.build_raw('profile.uuid', 'profile', 'personal_info',
                                                      external_raw_spec)
        non_external_spec = ColumnSpecification.build_raw('user.id', 'user', 'personal_info',
                                                          non_external_raw_spec)

        # then:
        self.assertEqual('profile.uuid', external_spec.field_name)
        self.assertEqual('profile', external_spec.object_type)
        self.assertEqual('personal_info', external_spec.main_category)
        # self.assertEqual(DataType.STRING, external_spec.data_type)
        self.assertTrue(external_spec.is_external_reference())

        # then:
        self.assertEqual('user.id', non_external_spec.field_name)
        self.assertFalse(non_external_spec.is_external_reference())

    def test_build_raw_spec_with_parent_spec(self):
        # given:
        raw_spec = {
            'value_type': 'boolean',
            'multivalue': True
        }

        # and:
        raw_single_value_parent_spec = {
            'multivalue': False
        }

        # and:
        raw_multi_value_parent_spec = {
            'multivalue': True
        }

        # when:
        single_column_spec = ColumnSpecification.build_raw('', '', '', raw_spec,
                                                           parent=raw_single_value_parent_spec)
        multi_column_spec = ColumnSpecification.build_raw('', '', '', raw_spec,
                                                          parent=raw_multi_value_parent_spec)

        # then:
        self.assertFalse(single_column_spec.is_field_of_list_element())
        self.assertTrue(multi_column_spec.is_field_of_list_element())

    def test_determine_converter_for_single_value(self):
        # expect:
        self._assert_correct_converter_single_value(DataType.STRING, StringConverter)
        self._assert_correct_converter_single_value(DataType.INTEGER, IntegerConverter)
        self._assert_correct_converter_single_value(DataType.BOOLEAN, BooleanConverter)
        self._assert_correct_converter_single_value(DataType.UNDEFINED, DefaultConverter)

    def _assert_correct_converter_single_value(self, data_type: DataType, expected_converter_type):
        # given:
        column_spec = ColumnSpecification('field', 'object_type', 'main_category', data_type)

        # when:
        converter = column_spec.determine_converter()

        # then:
        self.assertIsInstance(converter, expected_converter_type)

    def test_determine_converter_for_multivalue_type(self):
        # expect:
        self._assert_correct_converter_multivalue(DataType.STRING)
        self._assert_correct_converter_multivalue(DataType.INTEGER)
        self._assert_correct_converter_multivalue(DataType.BOOLEAN)
        self._assert_correct_converter_multivalue(DataType.UNDEFINED)

    def _assert_correct_converter_multivalue(self, data_type):
        # given:
        column_spec = ColumnSpecification('field', 'object_type', 'main_category',
                                          data_type, multivalue=True)

        # when:
        converter = column_spec.determine_converter()

        # then:
        self.assertIsInstance(converter, ListConverter)
        self.assertEqual(data_type, converter.base_type)

    def test_get_conversion_type_member_field(self):
        # given:
        column_spec = ColumnSpecification('user.name', 'user', 'user_data', DataType.STRING)

        # expect:
        self.assertEqual(ConversionType.MEMBER_FIELD, column_spec.get_conversion_type())

    def test_get_conversion_type_field_of_list_element(self):
        # given:
        column_spec = ColumnSpecification('product.product_name', 'product', 'store_entity',
                                          DataType.STRING, multivalue_parent=True)

        # expect:
        self.assertEqual(ConversionType.FIELD_OF_LIST_ELEMENT, column_spec.get_conversion_type())

    def test_get_conversion_type_identity(self):
        # given:
        column_spec = ColumnSpecification('product.product_id', 'product', 'store_entry',
                                          DataType.STRING, identity=True)

        # expect:
        self.assertEqual(ConversionType.IDENTITY, column_spec.get_conversion_type())

    def test_get_conversion_type_linked_identity(self):
        # given:
        column_spec = ColumnSpecification('account.number', 'user', 'profile_type',
                                          DataType.STRING, identity=True)

        # expect:
        self.assertEqual(ConversionType.LINKED_IDENTITY, column_spec.get_conversion_type())

    def test_get_conversion_type_linked_identity_order(self):
        # given:
        column_spec = ColumnSpecification('account.number', 'account', 'profile_type',
                                          DataType.STRING, identity=True, order_of_occurence=2)

        # expect:
        self.assertEqual(ConversionType.LINKED_IDENTITY, column_spec.get_conversion_type())

    def test_get_conversion_type_external_reference_identity(self):
        # given:
        column_spec = ColumnSpecification('account.uuid', 'user', 'profile_type',
                                          DataType.STRING, identity=True, external_reference=True)

        # expect:
        self.assertEqual(ConversionType.EXTERNAL_REFERENCE, column_spec.get_conversion_type())

    def test_get_conversion_type_linking_detail(self):
        # given:
        column_spec = ColumnSpecification('item.description', 'record', 'invoice_detail',
                                          DataType.STRING)

        # expect:
        self.assertEqual(ConversionType.LINKING_DETAIL, column_spec.get_conversion_type())



#!/usr/bin/env python


from ingest.template.schema_template import SchemaTemplate

schemas = [
    "https://schema.humancellatlas.org/type/project/5.1.0/project",
    "https://schema.humancellatlas.org/type/biomaterial/5.1.0/cell_suspension",
    "https://schema.humancellatlas.org/type/biomaterial/5.1.0/specimen_from_organism",
    "https://schema.humancellatlas.org/type/biomaterial/5.1.1/donor_organism",
    "https://schema.humancellatlas.org/type/file/5.1.0/sequence_file",
    "https://schema.humancellatlas.org/type/process/biomaterial_collection/5.1.0/collection_process",
    "https://schema.humancellatlas.org/type/process/biomaterial_collection/5.1.0/dissociation_process",
    "https://schema.humancellatlas.org/type/process/biomaterial_collection/5.1.0/enrichment_process",
    "https://schema.humancellatlas.org/type/process/sequencing/5.1.0/library_preparation_process",
    "https://schema.humancellatlas.org/type/process/sequencing/5.1.0/sequencing_process",
    "https://schema.humancellatlas.org/type/protocol/5.1.0/protocol",
    "https://schema.humancellatlas.org/type/protocol/biomaterial/5.1.0/biomaterial_collection_protocol",
    "https://schema.humancellatlas.org/type/protocol/sequencing/5.1.0/sequencing_protocol",
    "https://schema.humancellatlas.org/type/process/1.0.0/process",
]

template = SchemaTemplate(list_of_schema_urls=schemas)

# get key from user friendly name

# tabs = TabConfig().load("tabs_human_10x.yaml")

print (template.get_key_for_label("Biomaterial name", tab="Cell suspension"))

# lookup where to submit this entity

print (template.lookup("cell_suspension.schema.domain_entity"))

# lookup text field for donor_organism.human_specific.ethnicity.text

print (template.get_key_for_label("donor_organism.human_specific.ethnicity.text", tab="Donor organism"))



# lookup the schema url for project_core

print (template.lookup("project.project_core.schema.url"))

# get the user friendly name

print (template.lookup("project.project_core.project_title.user_friendly"))

# dump the config in yaml or json

print(template.yaml_dump(tabs_only=True))
# print(data.json_dump())



class ErrorReport:
    """
    A user friendly error message, along with corresponding ValidationError
    """
    def __init__(self, message="", validation_error=None, error_type=None):
        self.message = message
        self.validation_error = validation_error
        self.error_type = error_type

    def to_dict(self):
        error_report_dict = dict()

        error_report_dict["user_friendly_message"] = self.message

        error_report_dict["validation_error"] = dict()

        if self.error_type == "schema validation":
            error_report_dict["validation_error"]["absolute_path"] = list(self.validation_error.absolute_path)
            error_report_dict["validation_error"]["path"] = list(self.validation_error.path)
            error_report_dict["validation_error"]["message"] = self.validation_error.message
            error_report_dict["validation_error"]["instance"] = self.validation_error.instance
            error_report_dict["validation_error"]["schema_path"] = list(self.validation_error.schema_path)
            error_report_dict["validation_error"]["absolute_schema_path"] = list(self.validation_error.absolute_schema_path)
            error_report_dict["validation_error"]["validator"] = self.validation_error.validator
            error_report_dict["validation_error"]["validator_value"] = self.validation_error.validator_value
        else:
            error_report_dict["validation_error"]["absolute_path"] = self.validation_error.absolute_path
            error_report_dict["validation_error"]["path"] = self.validation_error.path
            error_report_dict["validation_error"]["message"] = self.validation_error.message
            error_report_dict["validation_error"]["instance"] = self.validation_error.instance

        return error_report_dict



from unittest import TestCase

from ingest.importer.conversion import utils


class ModuleTest(TestCase):

    def test_split_field_chain(self):
        # given:
        single = 'field'
        double = 'user.name'
        triple = 'user.address.city'

        # when:
        single_parent, single_target = utils.split_field_chain(single)
        double_parent, double_target = utils.split_field_chain(double)
        triple_parent, triple_target = utils.split_field_chain(triple)

        # then:
        self.assertEqual('', single_parent)
        self.assertEqual('field', single_target)

        # and:
        self.assertEqual('user', double_parent)
        self.assertEqual('name', double_target)

        # and:
        self.assertEqual('user.address', triple_parent)
        self.assertEqual('city', triple_target)

    def test_extract_root_field(self):
        # given:
        single = 'user'
        double = 'account.number'
        triple = 'product.item.id'

        # expect:
        self.assertEqual('user', utils.extract_root_field(single))
        self.assertEqual('account', utils.extract_root_field(double))
        self.assertEqual('product', utils.extract_root_field(triple))
        self.assertEqual('', utils.extract_root_field(''))
        self.assertIsNone(utils.extract_root_field(None))



#!/usr/bin/env python
"""
Given a tabs template and list of schema URLs, will output a spreadsheet in Xls format
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "08/05/2018"

from ingest.template import schema_template, tabs
from ingest.template.tabs import TabConfig
import xlsxwriter


class SpreadsheetBuilder:
    def __init__(self, output_file):

        self.workbook = xlsxwriter.Workbook(output_file)

        self.header_format = self.workbook.add_format({'bold': True, 'bg_color': '#D0D0D0'})
        self.locked_format = self.workbook.add_format({'locked': True})
        self.required_header_format = self.workbook.add_format({'bold': True, 'bg_color': '#FFFF00'})
        self.desc_format = self.workbook.add_format({'font_color': '#808080', 'italic': True, 'text_wrap': True})
        self.include_schemas_tab = False

    def generate_workbook(self, tabs_template=None, schema_urls=list(), include_schemas_tab=False):

        self.include_schemas_tab = include_schemas_tab
        if tabs_template:

            tabs_parser = TabConfig()
            tabs = tabs_parser.load(tabs_template)
            template = schema_template.SchemaTemplate(list_of_schema_urls=schema_urls, tab_config=tabs)
        else:
            template = schema_template.SchemaTemplate(list_of_schema_urls=schema_urls)

        self._build(template)
        return self

    def _get_value_for_column(self, template, col_name, property):
        try:
            uf = str(template.lookup(col_name + "."+property)) if template.lookup(col_name + "."+property) else ""
            return uf
        except:
            print("No property for " + col_name)
            return ""

    def get_user_friendly(self, template, col_name):
        key = col_name + ".user_friendly"
        try:
            uf = str(template.lookup(key)) if template.lookup(key) else col_name
            return uf
        except:
            return key

    def save_workbook(self):
        self.workbook.close()

    def _write_schemas(self, schema_urls):
        worksheet = self.workbook.add_worksheet("Schemas")
        worksheet.write(0, 0, "Schemas")
        for index, url in enumerate(schema_urls):
            worksheet.write(index + 1, 0, url)

    def _build(self, template):

        tabs = template.get_tabs_config()

        for tab in tabs.lookup("tabs"):

            for tab_name, detail in tab.items():

                worksheet = self.workbook.add_worksheet(detail["display_name"])

                col_number = 0

                for cols in detail["columns"]:

                    uf = self.get_user_friendly(template, cols)
                    desc = self._get_value_for_column(template, cols, "description")
                    required = bool(self._get_value_for_column(template, cols, "required"))
                    example_text = self._get_value_for_column(template, cols, "example")

                    hf = self.header_format
                    if required:
                        hf= self.required_header_format

                    # set the description
                    worksheet.write(0, col_number, desc, self.desc_format)

                    # set the user friendly name
                    worksheet.write(1, col_number, uf, hf)
                    worksheet.set_column(col_number, col_number, len(uf))

                    # write example
                    worksheet.write(2, col_number, example_text)

                    # set the key
                    worksheet.write(3, col_number, cols, self.locked_format)

                    col_number+=1

                worksheet.merge_range(first_col=0, first_row=4, last_col= len(detail["columns"]), last_row=4, cell_format= self.header_format, data="Add your data below this line")

        if self.include_schemas_tab:
            self._write_schemas(template.get_schema_urls())

        return self





import json
import os
import copy
import unittest
import uuid

import requests

from mock import MagicMock
from mock import Mock
from mock import patch


from unittest import TestCase

from ingest.exporter.ingestexportservice import IngestExporter
import ingest.exporter.ingestexportservice as ingestexportservice
import ingest.api.stagingapi as stagingapi

BASE_PATH = os.path.dirname(__file__)


class TestExporter(TestCase):
    def setUp(self):
        self.longMessage = True
        pass

    @patch('ingest.api.dssapi.DssApi')
    def test_get_input_bundle(self, dss_api_constructor):
        # given:
        dss_api_constructor.return_value = MagicMock('dss_api')

        # and:
        exporter = IngestExporter()

        # and:
        exporter.ingest_api.getRelatedEntities = MagicMock(return_value=['bundle1', 'bundle2'])
        process = {}

        # when:
        input_bundle = exporter.get_input_bundle(process)

        # then:
        self.assertEqual('bundle1', input_bundle)

    @unittest.skip
    @patch('ingest.api.dssapi.DssApi')
    def test_upload_metadata_files(self, dss_api_constructor):
        # given:
        dss_api_constructor.return_value = MagicMock('dss_api')

        # and:
        exporter = IngestExporter()

        # and:
        file_desc = stagingapi.FileDescription('checksums', 'contentType', 'name', 'name', 'file_url')
        exporter.upload_file = MagicMock(return_value=file_desc)
        metadata_files_info = {
            'project': {
                'dss_filename': 'project.json',
                'dss_uuid': 'uuid',
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'biomaterial': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'process': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'protocol': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'file': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'links': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            }
        }

        # when:
        metadata_files = exporter.upload_metadata_files('sub_uuid', metadata_files_info)

        # then:
        for metadata_list in metadata_files_info.values():
            for metadata_file in metadata_list:
                self.assertEqual(metadata_file['dss_uuid'], 'uuid')
                self.assertEqual(metadata_file['upload_file_url'], 'file_url')

    @patch('ingest.api.dssapi.DssApi')
    def test_upload_metadata_files_error(self, dss_api_constructor):
        # given:
        dss_api_constructor.return_value = MagicMock('dss_api')

        # and:
        exporter = IngestExporter()

        # and:
        exporter.upload_file = Mock(side_effect=Exception('test upload file error'))
        metadata_files_info = {
            'project': {
                'dss_filename': 'project.json',
                'dss_uuid': 'uuid',
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'biomaterial': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'process': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'protocol': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'file': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            },
            'links': {
                'dss_uuid': None,
                'content': {},
                'content_type': 'type',
                'upload_filename': 'filename'
            }
        }

        # when, then:
        with self.assertRaises(ingestexportservice.BundleFileUploadError) as e:
            metadata_files = exporter.upload_metadata_files('sub_uuid', metadata_files_info)

    @patch('ingest.api.dssapi.DssApi')
    def test_put_bundle_in_dss_error(self, dss_api_constructor):
        # given:
        dss_api_constructor.return_value = MagicMock('dss_api')

        # and:
        exporter = IngestExporter()

        # and:
        exporter.dss_api.put_bundle = Mock(side_effect=Exception('test create bundle error'))

        # when, then:
        with self.assertRaises(ingestexportservice.BundleDSSError) as e:
            metadata_files = exporter.put_bundle_in_dss('bundle_uuid', [])

    # mocks linked entities in the ingest API, attempts to build a bundle by crawling from an assay
    # process, asserts that the bundle created is equivalent to a known bundle
    @unittest.skip
    @patch('ingest.api.dssapi.DssApi')
    def test_create_bundle_manifest(self, dss_api_constructor):
        # given:
        dss_api_constructor.return_value = MagicMock('dss_api')

        class MockRequestResponse:
            def __init__(self, json_payload, status_code):
                self.payload = json_payload
                self.status_code = status_code
                self.text = json.dumps(json_payload)

            def json(self):
                return self.payload


        def mock_entities_url_to_file_dict():

            mock_entity_url_to_file_dict = dict()

            # analysis process
            mock_entity_url_to_file_dict["processes/mock-analysis-process-id"] = "/processes/mock_analysis_process.json"
            mock_entity_url_to_file_dict["processes/mock-analysis-process-id/derivedFiles"] = "/processes/mock_analysis_process_derived_files.json"
            mock_entity_url_to_file_dict["processes/mock-analysis-process-id/inputFiles"] = "/processes/mock_analysis_process_input_files.json"
            mock_entity_url_to_file_dict["processes/mock-analysis-process-id/inputBundleManifests"] = "/processes/mock_analysis_process_input_bundle_manifests.json"

            # input bundle manifests
            mock_entity_url_to_file_dict["bundleManifests/mock-input-bundle-manifest-id"] = "/processes/mock_bundle_manifest.json"

            # files
            mock_entity_url_to_file_dict["files/mock-fastq-read1-id"] = "/processes/mock_fastq_read1.json"
            mock_entity_url_to_file_dict["files/mock-fastq-read1-id/derivedByProcesses"] = "/files/mock_fastq_read1_derived_by_processes.json"
            mock_entity_url_to_file_dict["files/mock-fastq-read2-id"] = "/processes/mock_fastq_read2.json"
            mock_entity_url_to_file_dict["files/mock-fastq-read2-id/derivedByProcesses"] = "/files/mock_fastq_read2_derived_by_processes.json"

            # wrapper process(lib prep -> sequencing)
            mock_entity_url_to_file_dict["processes/mock-assay-process-id"] = "/processes/wrapper_process_lib_prep_and_sequencing.json"
            mock_entity_url_to_file_dict["processes/mock-assay-process-id/chainedProcesses"] = "/processes/wrapper_process_lib_prep_and_sequencing_chained_processes.json"
            mock_entity_url_to_file_dict["processes/mock-assay-process-id/inputBiomaterials"] = "/processes/wrapper_process_lib_prep_and_sequencing_input_biomaterial.json"
            mock_entity_url_to_file_dict["processes/mock-assay-process-id/derivedFiles"] = "/processes/wrapper_process_lib_prep_and_sequencing_derived_files.json"

            # lib prep process
            mock_entity_url_to_file_dict["processes/mock-lib-prep-process-id"] = "/processes/mock_lib_prep_process.json"
            mock_entity_url_to_file_dict["processes/mock-lib-prep-process-id/protocols"] = "/processes/mock_lib_prep_process_protocols.json"

            # sequencing process
            mock_entity_url_to_file_dict["processes/mock-sequencing-process-id"] = "/processes/mock_sequencing_process.json"
            mock_entity_url_to_file_dict["processes/mock-sequencing-process-id/protocols"] = "/processes/mock_sequencing_process_protocols.json"

            # cell suspension
            mock_entity_url_to_file_dict["biomaterials/mock-cell-suspension-id"] = "/biomaterials/mock_cell_suspension.json"
            mock_entity_url_to_file_dict["biomaterials/mock-cell-suspension-id/derivedByProcesses"] = "/biomaterials/mock_cell_suspension_derived_by_processes.json"

            # wrapper process(dissociation -> enrichment)
            mock_entity_url_to_file_dict["processes/mock-dissociation-enrichment-process-id"] = "/processes/wrapper_process_dissociation_and_enrichment.json"
            mock_entity_url_to_file_dict["processes/mock-dissociation-enrichment-process-id/chainedProcesses"] = "/processes/wrapper_process_dissociation_and_enrichment_chained_processes.json"
            mock_entity_url_to_file_dict["processes/mock-dissociation-enrichment-process-id/inputBiomaterials"] = "/processes/wrapper_process_dissociation_and_enrichment_input_biomaterial.json"
            mock_entity_url_to_file_dict["processes/mock-dissociation-enrichment-process-id/derivedBiomaterials"] = "/processes/wrapper_process_dissociation_and_enrichment_derived_biomaterial.json"

            # dissociation process
            mock_entity_url_to_file_dict["processes/mock-dissociation-process-id"] = "/processes/mock_dissociation_process.json"
            mock_entity_url_to_file_dict["processes/mock-dissociation-process-id/protocols"] = "/processes/mock_dissociation_process_protocols.json"

            # enrichment process
            mock_entity_url_to_file_dict["processes/mock-enrichment-process-id"] = "/processes/mock_encrichment_process.json"
            mock_entity_url_to_file_dict["processes/mock-enrichment-process-id/protocols"] = "/processes/mock_enrichment_process_protocols.json"

            # specimen
            mock_entity_url_to_file_dict["biomaterials/mock-specimen-id"] = "/biomaterials/mock_specimen.json"
            mock_entity_url_to_file_dict["biomaterials/mock-specimen-id/derivedByProcesses"] = "/biomaterials/mock_specimen_derived_by_processes.json"

            # sampling process
            mock_entity_url_to_file_dict["processes/mock-sampling-process-id"] = "/processes/mock_sampling_process.json"
            mock_entity_url_to_file_dict["processes/mock-sampling-process-id/inputBiomaterials"] = "/processes/mock_sampling_process_input_biomaterial.json"
            mock_entity_url_to_file_dict["processes/mock-sampling-process-id/derivedBiomaterials"] = "/processes/mock_sampling_process_derived_biomaterials.json"

            # donor
            mock_entity_url_to_file_dict["biomaterials/mock-donor-id"] = "/biomaterials/mock_donor.json"

            # project
            mock_entity_url_to_file_dict["projects/mock-project-id"] = "/projects/mock_project.json"

            return mock_entity_url_to_file_dict

        regular_requests_get = copy.deepcopy(requests.get)

        def mock_entities_retrieval(*args, **kwargs):
            test_ingest_dir = BASE_PATH + '/bundles/ingest-data'
            mock_entity_url_to_file_dict = mock_entities_url_to_file_dict()
            url = args[0]
            if 'mock-ingest-api' not in url:
                return regular_requests_get(*args, **kwargs)
            else: # mockville
                entity_relative_url = url.replace('http://mock-ingest-api/', '')
                if entity_relative_url in mock_entity_url_to_file_dict:
                    entity_file_location = mock_entity_url_to_file_dict[entity_relative_url]
                    with open(test_ingest_dir + entity_file_location, 'rb') as entity_file:
                        entity_json = json.load(entity_file)
                else: # don't have a mock for this entity; if it's a request for an empty input biomaterials/files/protocols, return a suitable empty _embedded
                    entity_json = {'_embedded' : dict(),
                                   '_links' :{
                                       'self' : {
                                           'href' : url
                                       }
                                   }}

                    if 'derivedByProcesses' in entity_relative_url or 'chainedProcesses' in entity_relative_url:
                        entity_json['_embedded'] = {'processes' : list()}
                    elif 'inputBiomaterials' in entity_relative_url or 'derivedBiomaterials' in entity_relative_url:
                        entity_json['_embedded'] = {'biomaterials' : list()}
                    elif 'inputBundleManifests' in entity_relative_url:
                        entity_json['_embedded'] = {'bundleManifests' : list()}
                    elif 'inputFiles' in entity_relative_url or 'derivedFiles' in entity_relative_url:
                        entity_json['_embedded'] = {'files' : list()}
                    elif 'protocols' in entity_relative_url:
                        entity_json['_embedded'] = {'protocols' : list()}
                    elif 'projects' in entity_relative_url:
                        with open(test_ingest_dir + '/projects/mock_project.json', 'rb') as project_file:
                            entity_json['_embedded'] = {'projects' : [json.load(project_file)]}
                    else:
                        raise Exception("Unknown resource in mock entities tests:" + url)

                return MockRequestResponse(entity_json, 200)

        # mock the calls to the ingest API for the entities in the bundle
        get_requests_mock = Mock()
        get_requests_mock.side_effect = mock_entities_retrieval
        requests.get = get_requests_mock

        exporter = ingestexportservice.IngestExporter()
        process_info = exporter.get_all_process_info('http://mock-ingest-api/processes/mock-assay-process-id')
        metadata_by_type = exporter.get_metadata_by_type(process_info)
        bundle_metadata_info = exporter.prepare_metadata_files(metadata_by_type)

        # assert that the contents of the bundle metadata info match that of the expected bundle
        self.assertEqual( # biomaterials...
            frozenset([biomaterial['hca_ingest']['document_id'] for biomaterial in bundle_metadata_info['biomaterial']['content']['biomaterials']]),
            frozenset([biomaterial['hca_ingest']['document_id'] for biomaterial in json_from_expected_bundle_file('assay/expected/Mouse Melanoma_biomaterial_bundle.json')['biomaterials']]))

        self.assertEqual( # processes...
            frozenset([process['hca_ingest']['document_id'] for process in bundle_metadata_info['process']['content']['processes']]),
            frozenset([process['hca_ingest']['document_id'] for process in json_from_expected_bundle_file('assay/expected/Mouse Melanoma_process_bundle.json')['processes']]))

        self.assertEqual( # protocols...
            frozenset([protocol['hca_ingest']['document_id'] for protocol in bundle_metadata_info['protocol']['content']['protocols']]),
            frozenset([protocol['hca_ingest']['document_id'] for protocol in json_from_expected_bundle_file('assay/expected/Mouse Melanoma_protocol_bundle.json')['protocols']]))

        self.assertEqual( # files...
            frozenset([file['hca_ingest']['document_id'] for file in bundle_metadata_info['file']['content']['files']]),
            frozenset([file['hca_ingest']['document_id'] for file in json_from_expected_bundle_file('assay/expected/Mouse Melanoma_file_bundle.json')['files']]))

        self.assertEqual( # links...
            frozenset([tuple(sorted(link.items())) for link in bundle_metadata_info['links']['content']['links']]),
            frozenset([tuple(sorted(link.items())) for link in json_from_expected_bundle_file('assay/expected/Mouse Melanoma_links_bundle.json')['links']])
        )

        self.assertEqual( # projects...
            bundle_metadata_info['project']['content']['hca_ingest']['document_id'],
            json_from_expected_bundle_file('assay/expected/Mouse Melanoma_project_bundle.json')['hca_ingest']['document_id']
        )

        # now run it on analysis
        process_info = exporter.get_all_process_info('http://mock-ingest-api/processes/mock-analysis-process-id')
        bundle_metadata_info = exporter.prepare_metadata_files(process_info)

        # assert that the contents of the bundle metadata info match that of the expected bundle
        self.assertEqual( # biomaterials...
            frozenset([biomaterial['hca_ingest']['document_id'] for biomaterial in bundle_metadata_info['biomaterial']['content']['biomaterials']]),
            frozenset([biomaterial['hca_ingest']['document_id'] for biomaterial in json_from_expected_bundle_file('analysis/expected/Mouse Melanoma_biomaterial_bundle.json')['biomaterials']]))

        self.assertEqual( # processes...
            frozenset([process['hca_ingest']['document_id'] for process in bundle_metadata_info['process']['content']['processes']]),
            frozenset([process['hca_ingest']['document_id'] for process in json_from_expected_bundle_file('analysis/expected/Mouse Melanoma_process_bundle.json')['processes']]))

        self.assertEqual( # protocols...
            frozenset([protocol['hca_ingest']['document_id'] for protocol in bundle_metadata_info['protocol']['content']['protocols']]),
            frozenset([protocol['hca_ingest']['document_id'] for protocol in json_from_expected_bundle_file('analysis/expected/Mouse Melanoma_protocol_bundle.json')['protocols']]))

        self.assertEqual( # files...
            frozenset([file['hca_ingest']['document_id'] for file in bundle_metadata_info['file']['content']['files']]),
            frozenset([file['hca_ingest']['document_id'] for file in json_from_expected_bundle_file('analysis/expected/Mouse Melanoma_file_bundle.json')['files']]))

        self.assertEqual( # links...
            frozenset([tuple(sorted(link.items())) for link in bundle_metadata_info['links']['content']['links']]),
            frozenset([tuple(sorted(link.items())) for link in json_from_expected_bundle_file('analysis/expected/Mouse Melanoma_links_bundle.json')['links']])
        )

        self.assertEqual( # projects...
            bundle_metadata_info['project']['content']['hca_ingest']['document_id'],
            json_from_expected_bundle_file('analysis/expected/Mouse Melanoma_project_bundle.json')['hca_ingest']['document_id']
        )


    def _create_entity_template(self):
        return {
            'submissionDate': '2018-03-14T09:53:02Z',
            'updateDate': '2018-03-14T09:53:02Z',
            'content': {},
            '_links': [],
            'events': [],
            'validationState': 'valid',
            'validationErrors': [],
            'uuid': {
                'uuid': '4674424e-3ab1-491c-8295-a68c7bb04b61'
            },
            'accession': 'accession123'
        }


def json_from_expected_bundle_file(relative_dir):
    # relative dir is relative to test/bundles
    with open(BASE_PATH + '/bundles/' + relative_dir, 'rb') as expected_bundle_file:
        return json.load(expected_bundle_file)



from unittest import TestCase

from ingest.importer.conversion.metadata_entity import MetadataEntity


class MetadataEntityTest(TestCase):

    def test_define_content(self):
        # given:
        metadata = MetadataEntity()

        # when:
        metadata.define_content('user.name', 'Juan dela Cruz')
        metadata.define_content('user.age', 47)

        # then:
        user = metadata.get_content('user')
        self.assertIsNotNone(user)
        self.assertEqual('Juan dela Cruz', user.get('name'))
        self.assertEqual(47, user.get('age'))

    def test_add_links(self):
        # given:
        metadata = MetadataEntity()

        # when:
        new_links = ['73f909', '83fddf1', '9004811']
        metadata.add_links('profile', new_links)

        # then:
        profile_links = metadata.get_links('profile')
        self.assertIsNotNone(profile_links)
        self.assertCountEqual(new_links, profile_links)

    def test_add_links_multiple_times(self):
        # given:
        metadata = MetadataEntity()

        # and:
        batch_1 = ['78de112', '963fefed']
        batch_2 = ['2daab01']

        # when:
        metadata.add_links('item', batch_1)
        metadata.add_links('item', batch_2)

        # then:
        item_links = metadata.get_links('item')
        self.assertIsNotNone(item_links)
        self.assertCountEqual(batch_1 + batch_2, item_links)

    def test_add_external_links(self):
        # given:
        metadata = MetadataEntity()

        # when:
        new_links = ['77701ee', '254aefb']
        metadata.add_external_links('file', new_links)

        # then:
        file_links = metadata.get_external_links('file')
        self.assertIsNotNone(file_links)
        self.assertCountEqual(new_links, file_links)

    def test_add_external_links_multiple_times(self):
        # given:
        metadata = MetadataEntity()

        # and:
        batch_1 = ['f8ede49']
        batch_2 = ['3dbb8b8']
        batch_3 = ['c23c45c']

        # when:
        metadata.add_external_links('process', batch_1)
        metadata.add_external_links('process', batch_2)
        metadata.add_external_links('process', batch_3)

        # then:
        process_links = metadata.get_external_links('process')
        self.assertIsNotNone(process_links)
        self.assertCountEqual(batch_2 + batch_1 + batch_3, process_links)

    def test_define_linking_detail(self):
        # given:
        metadata = MetadataEntity()

        # when:
        metadata.define_linking_detail('product_core.name', 'Apple Juice')
        metadata.define_linking_detail('product_core.description', 'pasteurised fruit juice')

        # then:
        product_core = metadata.get_linking_detail('product_core')
        self.assertIsNotNone(product_core)
        self.assertEqual('Apple Juice', product_core.get('name'))
        self.assertEqual('pasteurised fruit juice', product_core.get('description'))



#!/usr/bin/env python
"""
This package will return a SchemaTemplate objects
from a set of JSON schema files.
"""

__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "01/05/2018"

from datetime import datetime
from yaml import dump as yaml_dump
from yaml import load as yaml_load
from ingest.utils import doctict
from ingest.template.tabs import TabConfig
from ingest.api.ingestapi import IngestApi
import json
import jsonref
import re
import urllib.request



class SchemaTemplate:
    """
    A schema template is a simplified view over
    JSON schema for the HCA metadata
    """
    def __init__(self, ingest_api_url=None, list_of_schema_urls=None, tab_config=None):

        # todo remove this hard coding to a default ingest API url
        self.ingest_api_url = ingest_api_url if ingest_api_url else "http://api.ingest.dev.data.humancellatlas.org"
        self._template = {
            "template_version" : "1.0.0",
            "created_date" : str(datetime.now()),
            "meta_data_properties" : {},
            "labels" : {},
            "tabs": []
        }
        self._parser = SchemaParser(self)

        if not list_of_schema_urls:
            list_of_schema_urls = self.get_latest_submittable_schemas(self.ingest_api_url)
            print ("Got schemas from ingest api " + "\n".join(list_of_schema_urls))

        self.schema_urls = list_of_schema_urls
        self._load(self.schema_urls)

        self._tab_config  = TabConfig(init=self._template)
        if tab_config:
            # override the default tab config if one is supplied
            self._tab_config = tab_config

    def get_schema_urls (self):
        return self.schema_urls

    def get_latest_submittable_schemas(self, ingest_api_url):
        ingest_api = IngestApi(url=ingest_api_url)
        urls = []
        for schema in ingest_api.get_schemas(high_level_entity="type", latest_only=True):
            url = schema["_links"]["json-schema"]["href"]
            urls.append(url)
        return urls


    def _load(self, list_of_schema_urls):
        """
        given a list of URLs to JSON schema files
        return a SchemaTemplate object
        """
        for uri in list_of_schema_urls:
                with urllib.request.urlopen(uri) as url:
                    data = {}
                    try:
                        data = json.loads(url.read().decode())
                    except:
                        print("Failed to read schema from " + uri)
                    self._parser._load_schema(data)
        return self

    def get_tabs_config(self, ):
        return self._tab_config

    def lookup(self, key):
        try:
            return self.get(self._template["meta_data_properties"], key)
        except:
            raise UnknownKeyException(
                "Can't map the key to a known JSON schema property: " + str(key))

    def get_template(self):
        return self._template["meta_data_properties"]

    def append_tab(self, tab_info):
        self._template["tabs"].append(tab_info)

    def append_column_to_tab(self, property_key):
        level_one = self._get_level_one(property_key)
        for i, tab in enumerate(self._template["tabs"]):
            if level_one in tab:
                self._template["tabs"][i][level_one]["columns"].append(property_key)


    def put (self, property, value):
        '''
        Add a property to the schema template
        :param property:
        :param value:
        :return: void
        '''
        self._template["meta_data_properties"][property] = value

    def set_label_mappings(self, dict):
        '''
        A dictionary of label to keys mapping
        :param label:
        :param key:
        :return: void
        '''
        self._template["labels"] = dict

    def yaml_dump(self, tabs_only=False):
        return yaml_dump(yaml_load(self.json_dump(tabs_only)), default_flow_style=False)

    def json_dump(self, tabs_only=False):
        if tabs_only:
            tabs = {"tabs" : self._template["tabs"]}
            return json.dumps(tabs, indent=4)
        return json.dumps(self._template, indent=4)

    def get_key_for_label(self, column, tab):

        try:
            tab_key = self._tab_config.get_key_for_label(tab)
            for column_key  in self._parser.key_lookup(column.lower()):
                if tab_key == self._get_level_one(column_key):
                    return column_key
        except:
            raise UnknownKeyException(
                "Can't map the key to a known JSON schema property: " + str(column))

    def _get_level_one(self, key):
        return key.split('.')[0]

    def get(self, d, keys):
        if "." in keys:
            key, rest = keys.split(".", 1)
            return self.get(d[key], rest)
        else:
            return d[keys]


class SchemaParser:
    """A schema parser provides functions for
    accessing objects in a JSON schema"""
    def __init__(self, template):

        # always ignore these
        self.properties_to_ignore = \
            ["describedBy", "schema_version", "schema_type"]

        self.schema_template = template

        self._required = []

        # todo identifiable should be in the schema - hard coded here for now
        self._identifiable = ["biomaterial_id", "process_id", "protocol_id", "file_name"]

        self._key_lookup = {}

    def _load_schema(self, json_schema):
        """load a JSON schema representation"""
        # use jsonrefs to resolve all $refs in json
        data = jsonref.loads(json.dumps(json_schema))
        return self.__initialise_template(data)

    def key_lookup(self, key):
        return self._key_lookup[key]

    def __initialise_template(self, data):

        self._collect_required_properties(data)

        property = self._extract_property(data)
        if not property.schema or "type" not in property.schema.high_level_entity:
            raise RootSchemaException(
                "Schema must start with a root submittable type schema")
        else:
            # as this is top level add a retrievable property for
            property.uuid = {'external_reference': True, 'identifiable' : True}

        # todo get tab display name from schema
        tab_display = property.schema.module[0].upper() + property.schema.module[1:].replace("_", " ")
        tab_info = {property.schema.module : {"display_name": tab_display, "columns" : []}}

        self.schema_template.append_tab(tab_info)
        self.schema_template.put(property.schema.module, property)

        self._recursive_fill_properties(property.schema.module, data)

        self.schema_template.set_label_mappings(self._key_lookup)
        return self.schema_template

    def _get_path(self, str1, str2):
        return ".".join([str1, str2.split('/')[0]])

    def _recursive_fill_properties(self, path, data):

        for property_name, property_block in self._get_schema_properties_from_object(data).items():
            self._collect_required_properties(property_block)

            new_path =  self._get_path(path, property_name)
            property = self._extract_property(property_block, property_name=property_name, key=new_path)
            doctict.put(self.schema_template.get_template(), new_path, property)
            self._recursive_fill_properties(new_path, property_block)

    def _collect_required_properties(self, data):
        if "required" in data:
            self._required = list(set().union(self._required, data["required"]))

    def _new_template (self):
        return {
            "multivalue": False,
            "format":None,
            "required" : False,
            "identifiable": False,
            "external_reference": False,
            "user_friendly" : None,
            "description": None,
            "example" : None,
            "value_type": "string"}

    def _extract_property(self, data, *args, **kwargs):

        dic = self._new_template()

        if "type" in data:
            dic["value_type"] = data["type"]
            if data["type"] == "array":
                items = data.get("items", {})
                dic["value_type"] = items.get('type', 'string')
                dic["multivalue"] = True

        schema = self._get_schema_from_object(data)

        if 'property_name' in kwargs:
            if kwargs.get('property_name') in self._required:
                dic["required"] = True

            if kwargs.get('property_name') in self._identifiable:
                dic["identifiable"] = True


        if 'key' in kwargs and "object" != dic["value_type"]:
            self.schema_template.append_column_to_tab(kwargs.get('key'))

        if schema:
            dic["schema"] = schema


        # put the user friendly to key in the lookup table
        if 'key' in kwargs:

            self._update_label_to_key_map(kwargs.get("key"), kwargs.get("key"))

            if "user_friendly" in data:
                dic["user_friendly"] = data["user_friendly"]
                self._update_label_to_key_map(data["user_friendly"], kwargs.get("key"))

        if "description" in data:
            dic["description"] = data["description"]

        if "format" in data:
            dic["format"] = data["format"]

        if "example" in data:
            dic["example"] = data["example"]

        return doctict.DotDict(dic)

    def _update_label_to_key_map(self, label, key ):
        values = []
        if label.lower() not in self._key_lookup:
            values =  [ key ]
        else:
            values = self._key_lookup[label.lower()]
            values.append(key)

        if key not in self._key_lookup:
            self._key_lookup[key] = [key]

        self._key_lookup[label.lower()] = list(set(values))


    def _get_schema_from_object(self, data):
        """
        given a JSON object get the id and work out the
        """
        if "items" in data:
            return self._get_schema_from_object(data["items"])

        url_key = None

        if '$id' in data:
            url_key = '$id'

        if 'id' in data:
            url_key = 'id'

        if url_key:
            url = data[url_key]
            schema = Schema().build()
            schema.domain_entity = self.get_domain_entity_from_url(url)
            schema.high_level_entity = self.get_high_level_entity_from_url(url)
            schema.module = self.get_module_from_url(url)
            schema.url = url
            return schema

        return None

    def get_high_level_entity_from_url(self, url):
        pattern = re.compile("http[s]?://[^/]*/([^/]*)/")
        match = pattern.search(url)
        return match.group(1)

    def get_domain_entity_from_url(self, url):
        pattern = re.compile("http[s]?://[^/]*/[^/]*/(?P<domain_entity>.*)/(((\d+\.)?(\d+\.)?(\*|\d+))|(latest))/.*")
        match = pattern.search(url)
        return match.group(1) if match else None

    def get_module_from_url(self, url):
        return url.rsplit('/', 1)[-1]

    def _get_schema_properties_from_object(self, object):

        if "items" in object and isinstance(object["items"], dict):
            return self._get_schema_properties_from_object(object["items"])

        if "properties" in object and isinstance(object["properties"], dict):
            keys_to_remove = set(self.properties_to_ignore).intersection(set(object["properties"].keys()))

            for unwanted_key in keys_to_remove:
                del object["properties"][unwanted_key]
            return object["properties"]
        return {}


class Schema:
    def __init__(self):
        self.dict = {}

    def build(self):
        self.dict = {
            "high_level_entity": None,
            "domain_entity": None,
            "module": None,
            "url": None,
        }
        return doctict.DotDict(self.dict)


class Error(Exception):
    """Base-class for all exceptions raised by this module."""


class RootSchemaException(Error):
    """When generating a template we have to start with root JSON objects"""

class UnknownKeyException(Error):
    """Can't map the key to a known property"""

if __name__ == '__main__':
    pass


import copy

from ingest.importer.data_node import DataNode


class MetadataEntity:

    def __init__(self, object_id=None, content={}, links={}, external_links={}, linking_details={}):
        self.object_id = object_id
        self._content = DataNode(defaults=copy.deepcopy(content))
        self._links = copy.deepcopy(links)
        self._external_links = copy.deepcopy(external_links)
        self._linking_details = DataNode(defaults=copy.deepcopy(linking_details))

    @property
    def content(self):
        return copy.deepcopy(self._content)

    def get_content(self, content_property):
        return self._content[content_property]

    def define_content(self, content_property, value):
        self._content[content_property] = value

    def define_linking_detail(self, link_property, value):
        self._linking_details[link_property] = value

    @property
    def linking_details(self):
        return self._linking_details.as_dict()

    def get_linking_detail(self, link_property):
        return self._linking_details[link_property]

    @property
    def links(self):
        return copy.deepcopy(self._links)

    def get_links(self, link_entity_type):
        return self._links.get(link_entity_type)

    def add_links(self, link_entity_type, new_links):
        self._do_add_links(self._links, link_entity_type, new_links)

    @property
    def external_links(self):
        return copy.deepcopy(self._external_links)

    def get_external_links(self, link_entity_type):
        return self._external_links.get(link_entity_type)

    def add_external_links(self, link_entity_type, new_links):
        self._do_add_links(self._external_links, link_entity_type, new_links)

    @staticmethod
    def _do_add_links(link_map, link_entity_type, new_links):
        existent_links = link_map.get(link_entity_type)
        if existent_links is None:
            existent_links = []
            link_map[link_entity_type] = existent_links
        existent_links.extend(new_links)



from unittest import TestCase

from ingest.importer.conversion.data_converter import (
    BooleanConverter, ListConverter, DataType, IntegerConverter,
    StringConverter)
from ingest.importer.conversion.exceptions import InvalidBooleanValue


class DataTypeTest(TestCase):

    def test_find_case_insensitive(self):
        # expect:
        self.assertEqual(DataType.STRING, DataType.find('string'))
        self.assertEqual(DataType.STRING, DataType.find('String'))
        self.assertEqual(DataType.STRING, DataType.find('STRING'))

    def test_find_not_found(self):
        # expect:
        self.assertEqual(DataType.UNDEFINED, DataType.find('does_not_exist'))
        self.assertEqual(DataType.UNDEFINED, DataType.find(None))


class StringConverterTest(TestCase):

    def test_convert(self):
        # given:
        converter = StringConverter()

        # expect:
        self.assertEqual('data', converter.convert('data'))
        self.assertEqual('278', converter.convert(278))
        self.assertEqual('True', converter.convert(True))
        self.assertEqual('3.1416', converter.convert(3.1416))


class IntegerConverterTest(TestCase):

    def test_convert_from_string(self):
        # given:
        converter = IntegerConverter()

        # expect:
        self.assertEqual(37, converter.convert('37'))
        self.assertEqual(190, converter.convert('190'))

    def test_convert_from_float(self):
        # given:
        converter = IntegerConverter()

        # expect:
        self.assertEqual(899, converter.convert(899.003))
        self.assertEqual(200934, converter.convert(200934.118))

    def test_convert_from_integer(self):
        # given:
        converter = IntegerConverter()

        # expect:
        self.assertEqual(755, converter.convert(755))
        self.assertEqual(89221, converter.convert(89221))


class BooleanConverterTest(TestCase):

    def test_convert_to_false(self):
        converter = BooleanConverter()
        self.assertFalse(converter.convert('false'))
        self.assertFalse(converter.convert('False'))
        self.assertFalse(converter.convert('no'))
        self.assertFalse(converter.convert('No'))

    def test_convert_to_true(self):
        converter = BooleanConverter()
        self.assertTrue(converter.convert('true'))
        self.assertTrue(converter.convert('True'))
        self.assertTrue(converter.convert('yes'))
        self.assertTrue(converter.convert('Yes'))

    def test_convert_invalid_value(self):
        converter = BooleanConverter()

        with self.assertRaises(InvalidBooleanValue) as context:
            converter.convert('yup')

        self.assertEqual('yup', context.exception.get_value())


class ListConverterTest(TestCase):

    def test_convert_to_string_list(self):
        converter = ListConverter()
        self.assertEqual(['apple', 'banana', 'carrot'], converter.convert('apple||banana||carrot'))
        self.assertEqual(['durian', 'eggfruit', 'fig'], converter.convert('durian||eggfruit||fig'))

    def test_convert_to_int_list(self):
        converter = ListConverter(data_type=DataType.INTEGER)
        self.assertEqual([1, 2, 3], converter.convert('1||2||3'))

    def test_convert_to_int_list_multiple(self):
        converter = ListConverter(data_type=DataType.INTEGER)
        self.assertEqual([1, 2, 3], converter.convert('1||2||3'))

    def test_convert_to_int_list_single(self):
        converter = ListConverter(data_type=DataType.INTEGER)
        self.assertEqual([9606], converter.convert(9606))



#!/usr/bin/env python

# example cli run:
# python cli.py --submissionEnvelopeUuid=21ec3581-abc2-4e1c-9fd4-abce09aa5c19 --processUuid=b3f91034-f0de-4fb0-a038-b011b9f231b6 --ingest=http://api.ingest.dev.data.humancellatlas.org --dry

import logging
import sys

from optparse import OptionParser

from ingest.exporter.ingestexportservice import IngestExporter

if __name__ == '__main__':
    format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    logging.basicConfig(format=format, stream=sys.stdout, level=logging.INFO)

    parser = OptionParser()
    parser.add_option("-e", "--submissionEnvelopeUuid",
                      help="Submission envelope UUID for which to generate the bundle")
    parser.add_option("-p", "--processUuid",
                      help="Process UUID")
    parser.add_option("-D", "--dry", help="do a dry run without submitting to ingest", action="store_true",
                      default=False)
    parser.add_option("-o", "--output", dest="output",
                      help="output directory where to dump json files submitted to ingest", metavar="FILE",
                      default=None)
    parser.add_option("-i", "--ingest", help="the URL to the ingest API")
    parser.add_option("-s", "--staging", help="the URL to the staging API")
    parser.add_option("-d", "--dss", help="the URL to the datastore service")
    parser.add_option("-l", "--log", help="the logging level", default='INFO')

    (options, args) = parser.parse_args()

    if not options.submissionEnvelopeUuid:
        print ("You must supply a Submission Envelope UUID")
        exit(2)

    if not options.processUuid:
        print ("You must supply a process UUID.")
        exit(2)

    # TODO must only ask which environment to use
    if not options.ingest:
        print ("You must the url of Ingest API.")
        exit(2)

    exporter = IngestExporter(options)
    exporter.export_bundle(options.submissionEnvelopeUuid, options.processUuid)



#!/usr/bin/env python
"""
desc goes here
"""
import requests

__author__ = "jupp"
__license__ = "Apache 2.0"


import json
import logging
import os
import uuid
import time
import polling

from urllib.parse import urljoin

import ingest.api.dssapi as dssapi
import ingest.api.ingestapi as ingestapi
import ingest.api.stagingapi as stagingapi
from requests.exceptions import HTTPError

DEFAULT_INGEST_URL = os.environ.get('INGEST_API', 'http://api.ingest.dev.data.humancellatlas.org')
DEFAULT_STAGING_URL = os.environ.get('STAGING_API', 'http://upload.dev.data.humancellatlas.org')
DEFAULT_DSS_URL = os.environ.get('DSS_API', 'http://dss.dev.data.humancellatlas.org')

BUNDLE_SCHEMA_BASE_URL = os.environ.get('BUNDLE_SCHEMA_BASE_URL', 'https://schema.humancellatlas.org')


# TODO shouldn't source from environment variables, must pass config or params instead, throw an error if not in config

class IngestExporter:
    def __init__(self, options=None):
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        self.logger = logging.getLogger(__name__)

        self.dryrun = options.dry if options and options.dry else False
        self.outputDir = options.output if options and options.output else None

        self.ingestUrl = options.ingest if options and options.ingest else os.path.expandvars(DEFAULT_INGEST_URL)

        self.stagingUrl = options.staging if options and options.staging else os.path.expandvars(DEFAULT_STAGING_URL)
        self.dssUrl = options.dss if options and options.dss else os.path.expandvars(DEFAULT_DSS_URL)

        self.staging_api = stagingapi.StagingApi()
        self.dss_api = dssapi.DssApi()
        self.ingest_api = ingestapi.IngestApi(self.ingestUrl)
        self.related_entities_cache = {}

    def export_bundle(self, submission_uuid, process_uuid):
        start_time = time.time()
        self.related_entities_cache = {}
        saved_bundle_uuid = None

        if not self.dryrun and not self.staging_api.hasStagingArea(submission_uuid):
            error_message = "Can't do export as no upload area has been created."
            raise NoUploadAreaFoundError(error_message)

        self.logger.info('Export bundle for process with UUID ' + process_uuid)

        self.logger.info('Retrieving all process information...')

        process = self.ingest_api.getEntityByUuid('processes', process_uuid)
        process_info = self.get_all_process_info(process)

        self.logger.info('Generating bundle files...')
        submission = self.ingest_api.getEntityByUuid('submissionEnvelopes', submission_uuid)
        is_indexed = submission['triggersAnalysis']

        metadata_by_type = self.get_metadata_by_type(process_info)
        files_by_type = self.prepare_metadata_files(metadata_by_type, process_info, is_indexed)

        links = self.bundle_links(process_info.links)
        links_file_uuid = str(uuid.uuid4())
        files_by_type['links'] = list()
        files_by_type['links'].append({
            'content': links,
            'content_type': '"metadata/{0}"'.format('links'),
            'indexed': is_indexed,
            'dss_filename': 'links.json',
            'dss_uuid': links_file_uuid,
            'upload_filename': 'links_' + links_file_uuid + '.json'
        })

        # restructure bundle manifest
        bundle_manifest = self.create_bundle_manifest(submission_uuid, files_by_type)

        self.logger.info('Generating bundle files...')

        if self.dryrun:
            self.logger.info('Export is using dry run mode.')
            self.logger.info('Dumping bundle files...')

            for metadata_type in ['project', 'biomaterial', 'process', 'protocol', 'file', 'links']:
                for metadata_doc in files_by_type[metadata_type]:
                    bundle_file = metadata_doc
                    filename = bundle_file['upload_filename']
                    content = bundle_file['content']
                    output_dir = self.outputDir if self.outputDir else bundle_manifest.bundleUuid
                    self.dump_to_file(json.dumps(content, indent=4), filename, output_dir=output_dir)

            self.logger.info('Dry run for bundle ' + bundle_manifest.bundleUuid)
            self.logger.info("Execution Time: %s seconds" % (time.time() - start_time))
        else:
            self.logger.info('Uploading metadata files...')
            self.upload_metadata_files(submission_uuid, files_by_type)

            metadata_files = self.get_metadata_files(files_by_type)
            data_files = self.get_data_files(metadata_by_type['file'])
            bundle_files = metadata_files + data_files

            bundle_manifest.dataFiles = list()
            bundle_manifest.dataFiles = [data_file['dss_uuid'] for data_file in data_files]
            self.logger.info('Saving files in DSS...')
            bundle_uuid = bundle_manifest.bundleUuid
            created_files = self.put_files_in_dss(bundle_uuid, bundle_files, process_info)

            # check all created files
            self.logger.info('Verifying if all files get successfully copied to DSS...')
            self.verify_files(created_files)

            self.logger.info('Saving bundle in DSS...')
            self.put_bundle_in_dss(bundle_uuid, created_files)

            self.logger.info('Saving bundle manifest...')
            self.ingest_api.createBundleManifest(bundle_manifest)

            saved_bundle_uuid = bundle_manifest.bundleUuid

            self.logger.info('Bundle ' + bundle_uuid + ' was successfully created!')
            self.logger.info("Execution Time: %s seconds" % (time.time() - start_time))

        return saved_bundle_uuid

    def get_metadata_by_type(self, process_info: 'ProcessInfo') -> dict:
        #  given a ProcessInfo, pull out all the metadata and return as a map of UUID->metadata documents
        simplified = dict()
        simplified['process'] = dict(process_info.derived_by_processes)
        simplified['biomaterial'] = dict(process_info.input_biomaterials)
        simplified['protocol'] = dict(process_info.protocols)
        simplified['file'] = dict(process_info.derived_files)
        simplified['file'].update(process_info.input_files)
        simplified['file'].update(process_info.supplementary_files)

        simplified['project'] = dict()
        simplified['project'][process_info.project['uuid']['uuid']] = process_info.project

        return simplified

    def get_all_process_info(self, process):
        process_info = ProcessInfo()
        process_info.input_bundle = self.get_input_bundle(process)

        process_info.project = self.get_project_info(process)

        if not process_info.project:  # get from input bundle
            project_uuid_lists = list(process_info.input_bundle['fileProjectMap'].values())

            if len(project_uuid_lists) == 0 and len(project_uuid_lists[0]) == 0:
                raise Error('Input bundle manifest has no list of project uuid.')  # very unlikely to happen

            project_uuid = project_uuid_lists[0][0]
            process_info.project = self.ingest_api.getProjectByUuid(project_uuid)

        self.recurse_process(process, process_info)

        if process_info.project:
            supplementary_files = self.ingest_api.getRelatedEntities('supplementaryFiles', process_info.project, 'files')
            for supplementary_file in supplementary_files:
                uuid = supplementary_file['uuid']['uuid']
                process_info.supplementary_files[uuid] = supplementary_file

        return process_info

    def get_project_info(self, process):
        projects = list(self.ingest_api.getRelatedEntities('projects', process, 'projects'))

        if len(projects) > 1:
            raise MultipleProjectsError('Can only be one project in bundle')

        # TODO add checking for project only on an assay process
        # TODO an analysis process may have no link to a project

        if len(projects) > 0:
            return projects[0]

        return None

    # get all related info of a process
    def recurse_process(self, process, process_info):
        uuid = process['uuid']['uuid']
        process_info.derived_by_processes[uuid] = process

        # get all derived by processes using input biomaterial and input files
        derived_by_processes = []

        # wrapper process has the links to input biomaterials and derived files to check if a process is an assay
        input_biomaterials = self.get_related_entities('inputBiomaterials', process, 'biomaterials')
        for input_biomaterial in input_biomaterials:
            uuid = input_biomaterial['uuid']['uuid']
            process_info.input_biomaterials[uuid] = input_biomaterial
            derived_by_processes.extend(
                self.get_related_entities('derivedByProcesses', input_biomaterial, 'processes'))

        input_files = self.get_related_entities('inputFiles', process, 'files')
        for input_file in input_files:
            uuid = input_file['uuid']['uuid']
            process_info.input_files[uuid] = input_file
            derived_by_processes.extend(
                self.get_related_entities('derivedByProcesses', input_file, 'processes'))

        derived_biomaterials = self.get_related_entities('derivedBiomaterials', process, 'biomaterials')
        derived_files = self.get_related_entities('derivedFiles', process, 'files')

        process_uuid = process['uuid']['uuid']

        protocols = self.get_related_entities('protocols', process, 'protocols')
        for protocol in protocols:
            uuid = protocol['uuid']['uuid']
            process_info.protocols[uuid] = protocol

        for derived_file in derived_files:
            uuid = derived_file['uuid']['uuid']
            process_info.derived_files[uuid] = derived_file

        if input_biomaterials:
            if derived_files:
                process_info.links.append({
                    'process': process_uuid,
                    'inputs': [input_biomaterial['uuid']['uuid'] for input_biomaterial in input_biomaterials],
                    'input_type': 'biomaterial',
                    'outputs': [derived_file['uuid']['uuid'] for derived_file in derived_files],
                    'output_type': 'file',
                    'protocols': [
                        {
                            'protocol_type': self.get_concrete_entity_type(protocol),
                            'protocol_id': protocol['uuid']['uuid']
                        } for protocol in protocols
                    ]
                })

            if derived_biomaterials:
                process_info.links.append({
                    'process': process_uuid,
                    'inputs': [input_biomaterial['uuid']['uuid'] for input_biomaterial in input_biomaterials],
                    'input_type': 'biomaterial',
                    'outputs': [derived_biomaterial['uuid']['uuid'] for derived_biomaterial in derived_biomaterials],
                    'output_type': 'biomaterial',
                    'protocols': [
                        {
                            'protocol_type': self.get_concrete_entity_type(protocol),
                            'protocol_id': protocol['uuid']['uuid']
                        } for protocol in protocols
                    ]
                })

        if input_files and derived_files:
            process_info.links.append({
                'process': process_uuid,
                'inputs': [input_file['uuid']['uuid'] for input_file in input_files],
                'input_type': 'file',
                'outputs': [derived_file['uuid']['uuid'] for derived_file in derived_files],
                'output_type': 'file',
                'protocols': [
                    {
                        'protocol_type': self.get_concrete_entity_type(protocol),
                        'protocol_id': protocol['uuid']['uuid']
                    } for protocol in protocols
                ]
            })

        for derived_by_process in derived_by_processes:
            self.recurse_process(derived_by_process, process_info)

    def get_related_entities(self, relationship, entity, entity_type):
        entity_uuid = entity['uuid']['uuid']

        if self.related_entities_cache.get(entity_uuid) and self.related_entities_cache.get(entity_uuid).get(relationship):
            return self.related_entities_cache.get(entity_uuid).get(relationship)

        related_entities = list(self.ingest_api.getRelatedEntities(relationship, entity, entity_type))

        if not self.related_entities_cache.get(entity_uuid):
            self.related_entities_cache[entity_uuid] = {}

        if not self.related_entities_cache.get(entity_uuid).get(relationship):
            self.related_entities_cache[entity_uuid][relationship] = []

        self.related_entities_cache[entity_uuid][relationship] = related_entities

        return related_entities

    def get_input_bundle(self, process):
        bundle_manifests = list(self.ingest_api.getRelatedEntities('inputBundleManifests', process, 'bundleManifests'))

        if len(bundle_manifests) > 0:
            return bundle_manifests[0]

        return None

    def prepare_metadata_files(self, metadata_info, process_info, is_indexed=True) -> 'dict':
        metadata_files_by_type = dict()

        for entity_type in ['biomaterial', 'file', 'project', 'protocol', 'process']:
            metadata_files_by_type[entity_type] = list()
            concrete_type_ctr = dict()
            for (metadata_uuid, doc) in metadata_info[entity_type].items():
                concrete_type = self.get_concrete_entity_type(doc)

                concrete_type_ctr[concrete_type] = 0 if concrete_type not in concrete_type_ctr else concrete_type_ctr[concrete_type] + 1

                file_name = '{0}_{1}.json'.format(concrete_type, concrete_type_ctr[concrete_type])
                upload_filename = '{0}_{1}.json'.format(concrete_type, metadata_uuid)

                prepared_doc = {
                    'content': self.bundle_metadata(doc, metadata_uuid),
                    'content_type': '"metadata/{0}"'.format(entity_type),
                    'indexed': is_indexed,
                    'dss_filename': file_name,
                    'dss_uuid': metadata_uuid,
                    'upload_filename': upload_filename,
                    'update_date': doc['updateDate'],
                    'is_from_input_bundle': self._is_from_input_bundle(entity_type, metadata_uuid, process_info.input_bundle)
                }

                metadata_files_by_type[entity_type].append(prepared_doc)

        return metadata_files_by_type

    def _is_from_input_bundle(self, entity_type, metadata_uuid, input_bundle):

        field = {
            'biomaterial': 'fileBiomaterialMap',
            'process': 'fileProcessMap',
            'file': 'fileFilesMap',
            'project': 'fileProjectMap',
            'protocol': 'fileProtocolMap',
        }

        return input_bundle and input_bundle[field[entity_type]].get(metadata_uuid)

    def bundle_metadata(self, metadata_doc, uuid):
        provenance_core = dict()
        provenance_core['document_id'] = uuid
        provenance_core['submission_date'] = metadata_doc['submissionDate']
        provenance_core['update_date'] = metadata_doc['updateDate']

        bundle_doc = metadata_doc['content']
        bundle_doc['provenance'] = provenance_core

        return bundle_doc

    def bundle_links(self, links):

        latest_schema = self.ingest_api.get_schemas(
            latest_only=True,
            high_level_entity='system',
            domain_entity='',
            concrete_entity='links'
        )

        schema_url = latest_schema[0]['_links']['json-schema']['href'] if latest_schema else None
        schema_version = latest_schema[0]['schemaVersion'] if latest_schema else None

        return {
            'describedBy': schema_url,
            'schema_type': 'link_bundle',
            'schema_version': schema_version,
            'links': links
        }

    def upload_metadata_files(self, submission_uuid, metadata_files_info):
        try:
            for metadata_type in ['project', 'biomaterial', 'process', 'protocol', 'file', 'links']:
                for metadata_doc in metadata_files_info[metadata_type]:
                    bundle_file = metadata_doc
                    filename = bundle_file['upload_filename']
                    content = bundle_file['content']
                    content_type = bundle_file['content_type']

                    if not bundle_file.get('is_from_input_bundle'):
                        uploaded_file = self.upload_file(submission_uuid, filename, content, content_type)
                        bundle_file['upload_file_url'] = uploaded_file.url
        except Exception as e:
            message = "An error occurred on uploading bundle files: " + str(e)
            raise BundleFileUploadError(message)

    def put_bundle_in_dss(self, bundle_uuid, created_files):
        try:
            created_bundle = self.dss_api.put_bundle(bundle_uuid, created_files)
        except Exception as e:
            message = 'An error occurred while putting bundle in DSS: ' + str(e)
            raise BundleDSSError(message)

        return created_bundle

    def put_files_in_dss(self, bundle_uuid, files_to_put, process_info):
        created_files = []

        for bundle_file in files_to_put:
            version = ''
            file_uuid = bundle_file["dss_uuid"]
            created_file = None
            input_data_files = [input_file['dataFileUuid'] for input_file in list(process_info.input_files.values())]

            try:
                # TODO if file is an input file, this file may already be in the data store, need to get the stored version
                # This assumes that the latest version is the file version in the input bundle, should be a safe assumption for now
                # Ideally, bundle manifest must store the file uuid and version and version must be retrieved from there

                # if metadata file , check is_from_input_bundle flag, if true, do not put file to DSS again
                if bundle_file.get('is_from_input_bundle') or file_uuid in input_data_files:
                    file_response = self.dss_api.head_file(bundle_file["dss_uuid"])
                    created_file = {
                        'version': file_response.headers['X-DSS-VERSION']
                    }
                else:
                    created_file = self.dss_api.put_file(bundle_uuid, bundle_file)

                version = created_file['version']
            except Exception as e:
                raise FileDSSError('An error occurred while putting file in DSS' + str(e))

            file_param = {
                "indexed": bundle_file["indexed"],
                "name": bundle_file["submittedName"],
                "uuid": file_uuid,
                "content-type": bundle_file["content-type"],
                "version": version
            }

            created_files.append(file_param)

        return created_files

    def verify_files(self, created_files):
        for created_file in created_files:
            try:
                polling.poll(
                    lambda: self._is_file_copied(created_file),
                    step=30,
                    timeout=1200  # 20 minutes
                )
            except polling.TimeoutException as te:
                self.logger.error(f'File {created_file["uuid"]}/{created_file["version"]} with name {created_file["name"]} takes too long to be copied.')
                raise
            self.logger.info(f'File {created_file["uuid"]}/{created_file["version"]} with name {created_file["name"]} is successfully copied!')

    def _is_file_copied(self, created_file):
        try:
            r = self.dss_api.head_file(created_file["uuid"], version=created_file["version"])
            return (r.status_code == requests.codes.ok) or (r.status_code == requests.codes.created)
        except Exception as e:
            return False

    def get_metadata_files(self, metadata_files_info):
        metadata_files = []

        for entity_type in ['biomaterial', 'file', 'project', 'protocol', 'process', 'links']:
            for metadata_file in metadata_files_info[entity_type]:
                metadata_files.append({
                    'name': metadata_file['upload_filename'],
                    'submittedName': metadata_file['dss_filename'],
                    'url': metadata_file.get('upload_file_url'),
                    'dss_uuid': metadata_file['dss_uuid'],
                    'indexed': metadata_file['indexed'],
                    'content-type': metadata_file['content_type'],
                    'update_date': metadata_file.get('update_date'),
                    'is_from_input_bundle': metadata_file.get('is_from_input_bundle')
                })
        return metadata_files

    def get_data_files(self, uuid_file_dict):
        data_files = []
        #  TODO: need to keep track of UUIDs used so that retries work when the DSS returns a 500
        for file_uuid, data_file in uuid_file_dict.items():
            filename = data_file['fileName']
            cloud_url = data_file['cloudUrl']
            data_file_uuid = data_file['dataFileUuid']

            data_files.append({
                'name': filename,
                'submittedName': filename,
                'url': cloud_url,
                'dss_uuid': data_file_uuid,
                'indexed': False,
                'content-type': 'data'
            })

        return data_files

    def create_bundle_manifest(self, submission_uuid, files_by_type):
        bundle_manifest = ingestapi.BundleManifest()
        bundle_manifest.envelopeUuid = submission_uuid

        bundle_manifest.fileProjectMap = dict()
        for metadata_file in files_by_type['project']:
            bundle_manifest.fileProjectMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileBiomaterialMap = dict()
        for metadata_file in files_by_type['biomaterial']:
            bundle_manifest.fileBiomaterialMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileProcessMap = dict()
        for metadata_file in files_by_type['process']:
            bundle_manifest.fileProcessMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileProtocolMap = dict()
        for metadata_file in files_by_type['protocol']:
            bundle_manifest.fileProtocolMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileFilesMap = dict()
        for metadata_file in files_by_type['file']:
            bundle_manifest.fileFilesMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        return bundle_manifest

    def get_concrete_entity_type(self, schema_uri):
        return schema_uri["content"]["describedBy"].rsplit('/', 1)[-1]

    def upload_file(self, submission_uuid, filename, content, content_type):
        file_description = self.staging_api.getFile(submission_uuid, filename)

        if file_description:
            self.logger.info(f"The file {filename} already exists in the Upload area {submission_uuid}.")
        else:
            self.logger.info("Writing to staging area..." + filename)
            try:
                file_description = self.staging_api.stageFile(submission_uuid, filename, content, content_type)
            except HTTPError as e:
                if str(e.response.status_code) == "409":
                    file_description = self.staging_api.getFile(submission_uuid, filename)
                    if file_description:
                        return file_description
                    else:
                        raise e
                raise e

        self.logger.info("File staged at " + file_description.url)
        return file_description

    def dump_to_file(self, content, filename, output_dir='output'):
        directory = os.path.abspath(output_dir)

        if not os.path.exists(directory):
            os.makedirs(directory)

        tmp_file = open(directory + "/" + filename + ".json", "w")
        tmp_file.write(content)
        tmp_file.close()

class File:
    def __init__(self):
        self.name = ""
        self.content_type = ""
        self.size = ""
        self.id = ""
        self.checksums = {}


class ProcessInfo:
    def __init__(self):
        self.project = {}

        # uuid => object mapping
        self.input_biomaterials = {}
        self.derived_by_processes = {}
        self.input_files = {}
        self.derived_files = {}
        self.protocols = {}
        self.supplementary_files = {}

        self.links = []

        self.input_bundle = None


# Module Exceptions


class Error(Exception):
    """Base-class for all exceptions raised by this module."""


class MultipleProjectsError(Error):
    """A process should only have one project linked."""


class InvalidBundleError(Error):
    """There was a failure in bundle validation."""

class BundleFileUploadError(Error):
    """There was a failure in bundle file upload."""


class BundleDSSError(Error):
    """There was a failure in bundle creation in DSS."""


class FileDSSError(Error):
    """There was a failure in file creation in DSS."""


class NoUploadAreaFoundError(Error):
    """Export couldn't be as no upload area found"""